# pyOCD debugger
# Copyright (c) 2023 David van Rijn
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...coresight.cortex_m import CortexM
from ...coresight.minimal_mem_ap import MinimalMemAP as MiniAP

LOG = logging.getLogger(__name__)

class DBGMCU:

    # Via APB-ap (AP2)
    #BASE =   0xe00e1000
    # via AHB-ap (AP0,1)
    BASE =   0x5c001000

    IDC =  BASE + 0x000
    CR =   BASE + 0x004
    CR_VALUE = (0x3f | # keep running in stop sleep and standby
               0x07 << 20 | # enable all debug components
               0x07
               )

    ABP3 = BASE + 0x034

class FlashPeripheral:
    def __init__(self):
        self.flashaddr = 0x2000+0x12000000+0x40000000
        self.flash_keyr = self.flashaddr + 4
        self.flash_optkeyr = self.flashaddr + 8
        self.flash_optcr = self.flashaddr + 0x18
        self.flash_cr = self.flashaddr + 0xc
        self.flash_sr = self.flashaddr + 0x10
        self.flash_optsr_cur = self.flashaddr + 0x1c
        self.flash_optsr_prg = self.flashaddr + 0x20


FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
        0xe7fdbe00,
        0x4603b510, 0x00def44f, 0x60e04c7f, 0x487ebf00, 0xf0006900, 0x28000004, 0x487cd1f9, 0x60604c7a,
        0x6060487b, 0xbd102000, 0x48774601, 0xf04068c0, 0x4a750001, 0x200060d0, 0xbf004770, 0x69004872,
        0x0001f000, 0xd1f92800, 0x486fbf00, 0xf0006900, 0x28000001, 0x486cd1f9, 0xf02068c0, 0x496a0030,
        0x460860c8, 0xf04068c0, 0x60c80008, 0x68c04608, 0x0080f040, 0xbf0060c8, 0x69004863, 0x0001f000,
        0xd1f92800, 0x68c04860, 0x0008f020, 0x60c8495e, 0x47702000, 0xf3c14601, 0xf1b14243, 0xd3376f00,
        0x6f01f1b1, 0xbf00d234, 0x69004857, 0x0004f000, 0xd1f92800, 0x68c04854, 0x60e0f420, 0x60d84b52,
        0x68c04618, 0x7330f647, 0x4b4f4398, 0x461860d8, 0x230468c0, 0x2302ea43, 0x4b4b4318, 0x461860d8,
        0xf04068c0, 0x60d80080, 0x4847bf00, 0xf0006900, 0x28000004, 0x4844d1f9, 0xf02068c0, 0x4b420004,
        0x461860d8, 0xf4006900, 0xb1080080, 0x47702001, 0xe7fc2000, 0x4603b5f0, 0x461c4616, 0x22004635,
        0x6f00f1b3, 0xf1b3d310, 0xd20d6f01, 0x4836bf00, 0xf0006900, 0x28000004, 0xbf00d1f9, 0x69004832,
        0x0001f000, 0xd1f92800, 0x00def44f, 0x60f84f2e, 0xf1b3e056, 0xd3536f00, 0x6f01f1b3, 0x482ad250,
        0xf64768c0, 0x43b87730, 0x60f84f27, 0x60f82002, 0xd30c2920, 0xe0062200, 0xf855686f, 0x60670b08,
        0x0b08f844, 0x2a041c52, 0x3920dbf6, 0x4620e017, 0x2200462f, 0xf817e004, 0xf800cb01, 0x1c52cb01,
        0xd3f8428a, 0xe0042200, 0x0cfff04f, 0xcb01f800, 0xf1c11c52, 0x45940c20, 0x2100d8f6, 0xbf00bf00,
        0x69004811, 0x0004f000, 0xd1f92800, 0x480ebf00, 0xf0006900, 0x28000001, 0x480bd1f9, 0xf4006900,
        0xb9683080, 0x6f00f1b3, 0xf1b3d308, 0xd2056f01, 0x68c04805, 0x0002f020, 0x60f84f03, 0xbdf02001,
        0xd1a62900, 0xe7fa2000, 0x52002000, 0x45670123, 0xcdef89ab, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000005,
    'pc_unInit': 0x2000002d,
    'pc_program_page': 0x20000119,
    'pc_erase_sector': 0x20000099,
    'pc_eraseAll': 0x2000003f,

    'static_base' : 0x20000000 + 0x00000004 + 0x00000214,
    'begin_stack' : 0x20001a20,
    'end_stack' : 0x20000a20,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x400,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    # Enable double buffering
    'page_buffers' : [
        0x20000220,
        0x20000620
    ],
    'min_program_length' : 0x400,

    # Relative region addresses and sizes
    'ro_start': 0x4,
    'ro_size': 0x214,
    'rw_start': 0x218,
    'rw_size': 0x4,
    'zi_start': 0x21c,
    'zi_size': 0x0,

    # Flash information
    'flash_start': 0x8000000,
    'flash_size': 0x100000,
    'sector_sizes': (
        (0x0, 0x20000),
    )
}



class STM32H723xx(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x08000000, length=0x100000, sector_size=0x8000,
                                                        page_size=0x400,
                                                        is_boot_memory=True,
                                                        algo=FLASH_ALGO),
        #ITCM
        RamRegion(   start=0x00000000, length=0x10000,
                  is_cachable=False,
                  access="rwx"),
        #DTCM
        RamRegion(   start=0x20000000, length=0x20000,
                  is_cachable=False,
                  access="rw"),
        #sram1
        RamRegion(   start=0x30000000, length=0x4000,
                  is_powered_on_boot=False),
        #sram2
        RamRegion(   start=0x30004000, length=0x4000,
                  is_powered_on_boot=False),
        #sram4
        RamRegion(   start=0x38000000, length=0x4000),
        )



    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)

    def assert_reset_for_connect(self):
        self.dp.assert_reset(1)

    def safe_reset_and_halt(self):
        assert self.dp.is_reset_asserted()

        # At this point we can't access full AP as it is not initialized yet.
        # Let's create a minimalistic AP and use it.
        ap = MiniAP(self.dp)
        ap.init()

        DEMCR_value = ap.read32(CortexM.DEMCR)

        # Halt on reset.
        ap.write32(CortexM.DEMCR,
                   CortexM.DEMCR_VC_CORERESET |
                   CortexM.DEMCR_TRCENA
                   )
        ap.write32(CortexM.DHCSR, CortexM.DBGKEY | CortexM.C_DEBUGEN)

        self.dp.assert_reset(0)
        time.sleep(0.01)

        DEV_ID = ap.read32(DBGMCU.IDC) & 0xfff
        assert DEV_ID == 0x483, f"IDC.DEV_ID 0x{DEV_ID:03x} did not match expected. 0x483"
        ap.write32(DBGMCU.CR, DBGMCU.CR_VALUE)

        CR = ap.read32(DBGMCU.CR)
        LOG.info("CR: 0x%08x", CR)

        # Restore DEMCR original value.
        ap.write32(CortexM.DEMCR, DEMCR_value)

    def create_init_sequence(self):
        # this was copied from target_STM32F767xx.py but seems to apply here as well
        #
        # STM32 under some low power/broken clock states doesn't allow AHP communication.
        # Low power modes are quite popular on stm32 (including MBed OS defaults).
        # 'attach' mode is broken by default, as STM32 can't be connected on low-power mode
        #  successfully without previous DBGMCU setup (It is not possible to write DBGMCU).
        # It is also not possible to run full pyOCD discovery code under-reset.
        #
        # As a solution we can setup DBGMCU under reset, halt core and release reset.
        # Unfortunately this code has to be executed _before_ discovery stage
        # and without discovery stage we don't have access to AP/Core.
        # As a solution we can create minimalistic AP implementation and use it
        # to setup core halt.
        # So the sequence for 'halt' connect mode will look like
        # -> Assert reset
        # -> Connect DebugPort
        # -> Setup MiniAp
        # -> Setup halt on reset
        # -> Enable support for debugging in low-power modes
        # -> Release reset
        # -> [Core is halted and reset is released]
        # -> Continue [discovery, create cores, etc]
        seq = super().create_init_sequence()
        if self.session.options.get('connect_mode') in ('halt', 'under-reset'):
            seq.insert_before('dp_init', ('assert_reset_for_connect', self.assert_reset_for_connect))
            seq.insert_after('dp_init', ('safe_reset_and_halt', self.safe_reset_and_halt))

        return seq

    def _unlock_flash_peripheral(self):
        bank = FlashPeripheral()
        LOG.info('unlocking flash peripheral')
        self.reset_and_halt()
        while self.read32(bank.flash_sr) & 1:
            time.sleep(0.1)

        if self.read32(bank.flash_cr) & 1 != 0:
            self.write32(bank.flash_keyr,    0x4567_0123)
            self.write32(bank.flash_keyr,    0xCDEF_89AB)
            while self.read32(bank.flash_sr) & 1:
                time.sleep(0.1)
        if self.read32(bank.flash_optcr) & 1 != 0:
            self.write32(bank.flash_optkeyr, 0x0819_2A3B)
            self.write32(bank.flash_optkeyr, 0x4C5D_6E7F)



    def is_locked(self):
        bank = FlashPeripheral()
        optsr = self.read32(bank.flash_optsr_prg)
        rdp = optsr & 0x0000_ff00
        if rdp == 0xaa:
            return False;
        if rdp == 0xcc:
            LOG.warning("MCU permanently locked. No unlock possible")
        return True

    def disable_read_protection(self):
        bank = FlashPeripheral()
        self._unlock_flash_peripheral()

        while self.read32(bank.flash_sr) & 1:
            time.sleep(0.1)

        optsr = self.read32(bank.flash_optsr_prg)
        self.write32(bank.flash_optsr_prg, optsr & 0xffff_00ff | 0x0000_aa00)
        self.write32(bank.flash_optcr, 2)
        while self.read32(bank.flash_sr) & 1:
            time.sleep(0.1)
        self.reset_and_halt()

    def mass_erase(self):
        bank = FlashPeripheral()
        self._unlock_flash_peripheral()

        while self.read32(bank.flash_sr) & 1:
            time.sleep(0.1)

        self.write32(bank.flash_cr, 1<<3 | 3<<4)
        self.write32(bank.flash_cr, 1<<3 | 3<<4 | 1<<7)
        LOG.info("mass_erase")
        while self.read32(bank.flash_sr) & 1:
            time.sleep(0.1)
        LOG.info("mass_erase done")








# pyOCD debugger
# Copyright (c) 2006-2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# Copyright (c) 2022 Harper Weigle
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import collections

from .interface import Interface
from .common import (
    filter_device_by_usage_page,
    generate_device_unique_id,
    is_known_cmsis_dap_vid_pid,
    )
from ..dap_access_api import DAPAccessIntf
from ....utility.timeout import Timeout

OPEN_TIMEOUT_S = 60.0

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

try:
    import pywinusb.hid as hid
except ImportError:
    IS_AVAILABLE = False
else:
    IS_AVAILABLE = True

class PyWinUSB(Interface):
    """@brief CMSIS-DAP USB interface class using pyWinUSB for the backend."""

    isAvailable = IS_AVAILABLE

    def __init__(self, dev):
        super().__init__()
        # Vendor page and usage_id = 2
        reports = dev.find_output_reports()
        assert len(reports) == 1
        self.report = reports[0]
        self.packet_size = len(self.report.get_raw_data()) - 1
        self.vendor_name = dev.vendor_name or f"{dev.vendor_id:#06x}"
        self.product_name = dev.product_name or f"{dev.product_id:#06x}"
        self.serial_number = dev.serial_number \
                or generate_device_unique_id(dev.vendor_id, dev.product_id, dev.device_path)
        self.vid = dev.vendor_id
        self.pid = dev.product_id
        self.device = dev
        # deque used here instead of synchronized Queue
        # since read speeds are ~10-30% faster and are
        # comparable to a list based implementation.
        self.rcv_data = collections.deque()

    # handler called when a report is received
    def rx_handler(self, data):
        if TRACE.isEnabledFor(logging.DEBUG):
            # Strip off trailing zero bytes to reduce clutter.
            TRACE.debug("  USB IN < (%d) %s", len(data), ' '.join([f'{i:02x}' for i in bytes(data).rstrip(b'\x00')]))

        self.rcv_data.append(data[1:])

    def open(self):
        self.device.set_raw_data_handler(self.rx_handler)

        # Attempt to open the device.
        # Note - this operation must be retried since
        # other instances of pyOCD listing board can prevent
        # opening this device with exclusive access.
        with Timeout(OPEN_TIMEOUT_S, sleeptime=0.25) as t_o:
            while t_o.check():
                # Attempt to open the device
                try:
                    self.device.open(shared=False)
                    break
                except hid.HIDError:
                    pass

                # Attempt to open the device in shared mode to make
                # sure it is still there
                try:
                    self.device.open(shared=True)
                    self.device.close()
                except hid.HIDError as exc:
                    # If the device could not be opened in read only mode
                    # Then it either has been disconnected or is in use
                    # by another thread/process
                    raise DAPAccessIntf.DeviceError(f"Unable to open device {self.serial_number}") from exc

            else:
                # If this timeout has elapsed then another process
                # has locked this device in shared mode. This should
                # not happen.
                raise DAPAccessIntf.DeviceError(f"Timed out attempting to open device {self.serial_number}")

    @staticmethod
    def get_all_connected_interfaces():
        """@brief Returns all the connected CMSIS-DAP devices"""
        all_devices = hid.find_all_hid_devices()

        # find devices with good vid/pid
        all_mbed_devices = []
        for d in all_devices:
            known_cmsis_dap = is_known_cmsis_dap_vid_pid(d.vendor_id, d.product_id)
            if ("CMSIS-DAP" in d.product_name) or known_cmsis_dap:
                all_mbed_devices.append(d)

        boards = []
        for dev in all_mbed_devices:
            try:
                dev.open(shared=True)

                # Perform device-specific filtering.
                if filter_device_by_usage_page(dev.vendor_id, dev.product_id, dev.hid_caps.usage_page):
                    dev.close()
                    continue

                report = dev.find_output_reports()
                if len(report) != 1:
                    dev.close()
                    continue
                new_board = PyWinUSB(dev)
                boards.append(new_board)
            except Exception as e:
                if (str(e) != "Failure to get HID pre parsed data"):
                    LOG.error("Receiving Exception: %s", e)
            finally:
                dev.close()

        return boards

    def write(self, data):
        """@brief Write data on the OUT endpoint associated to the HID interface"""
        if TRACE.isEnabledFor(logging.DEBUG):
            TRACE.debug("  USB OUT> (%d) %s", len(data), ' '.join([f'{i:02x}' for i in data]))

        data.extend([0] * (self.packet_size - len(data)))
        self.report.send([0] + data)

    def read(self):
        """@brief Read data on the IN endpoint associated to the HID interface"""
        # Spin for a while if there's not data available yet. 100 µs sleep between checks.
        with Timeout(self.DEFAULT_USB_TIMEOUT_S, sleeptime=0.0001) as t_o:
            while t_o.check():
                if len(self.rcv_data):
                    break
            else:
                # Read operations should typically take ~1-2ms.
                # If this exception occurs, then it could indicate
                # a problem in one of the following areas:
                # 1. Bad usb driver causing either a dropped read or write
                # 2. CMSIS-DAP firmware problem cause a dropped read or write
                # 3. CMSIS-DAP is performing a long operation or is being
                #    halted in a debugger
                raise DAPAccessIntf.DeviceError(f"Timeout reading from device {self.serial_number}")

        # Trace when the higher layer actually gets a packet previously read.
        if TRACE.isEnabledFor(logging.DEBUG):
            # Strip off trailing zero bytes to reduce clutter.
            TRACE.debug("  USB RD < (%d) %s", len(self.rcv_data[0]),
                    ' '.join([f'{i:02x}' for i in bytes(self.rcv_data[0]).rstrip(b'\x00')]))

        return self.rcv_data.popleft()

    def close(self):
        """@brief Close the interface"""
        LOG.debug("closing interface")
        self.device.close()



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

# to_bytes_safe() converts a unicode string to a bytes object by encoding as
# latin-1. It will also accept a value that is already a bytes object and
# return it unmodified.
def to_bytes_safe(v: Union[str, bytes]) -> bytes:
    if isinstance(v, str):
        return v.encode('utf-8')
    else:
        return v

# to_str_safe() converts a bytes object to a unicode string by decoding from
# latin-1. It will also accept a value that is already a str object and
# return it unmodified.
def to_str_safe(v: Union[str, bytes]) -> str:
    if isinstance(v, str):
        return v
    else:
        return v.decode('utf-8')




# pyOCD debugger
# Copyright (c) 2017-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .component import CoreSightComponent

class TPIU(CoreSightComponent):
    """@brief Trace Port Interface Unit"""

    # Register definitions.
    #
    # The addresses are offsets from the base address.
    ACPR = 0x00000010
    ACPR_PRESCALER_MASK = 0x0000ffff

    SPPR = 0x000000f0
    SPPR_TXMODE_MASK = 0x00000003
    SPPR_TXMODE_NRZ = 0x00000002

    FFCR = 0x00000304
    FFCR_ENFCONT_MASK = (1 << 1)

    DEVID = 0x00000fc8
    DEVID_NRZ_MASK = (1 << 11)

    def __init__(self, ap, cmpid=None, addr=None):
        """@brief Standard CoreSight component constructor."""
        super(TPIU, self).__init__(ap, cmpid, addr)
        self._has_swo_uart = False

    @property
    def has_swo_uart(self):
        """@brief Whether SWO UART mode is supported by the TPIU."""
        return self._has_swo_uart

    def init(self):
        """@brief Reads TPIU capabilities.

        Currently this method simply checks whether the TPIU supports SWO in asynchronous
        UART mode. The result of this check is available via the has_swo_uart property.
        """
        devid = self.ap.read32(self.address + TPIU.DEVID)
        self._has_swo_uart = (devid & TPIU.DEVID_NRZ_MASK) != 0

    def set_swo_clock(self, swo_clock, system_clock):
        """@brief Prepare TPIU for transmitting SWO at a given baud rate.

        Configures the TPIU for SWO UART mode, then sets the SWO clock frequency based on
        the provided system clock.

        @param self
        @param swo_clock Desired SWO baud rate in Hertz.
        @param system_clock The frequency of the SWO clock source in Hertz. This is almost always
            the system clock, also called the HCLK or fast clock.
        @return Boolean indicating if SWO UART mode could be configured with the requested
            baud rate set within 3%.
        """
        # First check whether SWO UART is supported.
        if not self.has_swo_uart:
            return False

        # Go ahead and configure for SWO.
        self.ap.write32(self.address + TPIU.SPPR, TPIU.SPPR_TXMODE_NRZ) # Select SWO UART mode.
        self.ap.write32(self.address + TPIU.FFCR, 0) # Disable formatter.

        # Compute the divider.
        div = (system_clock // swo_clock) - 1
        actual = system_clock // (div + 1)
        deltaPercent = abs(swo_clock - actual) / swo_clock
        # Make sure the target baud rate was met with 3%.
        if deltaPercent > 0.03:
            return False
        self.ap.write32(self.address + TPIU.ACPR, div & TPIU.ACPR_PRESCALER_MASK)
        return True





# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Optional, TYPE_CHECKING)

from .target import (Target, TargetGraphNode)

if TYPE_CHECKING:
    from ..debug.context import DebugContext
    from ..debug.elf.elf import ELFBinaryFile

class CoreTarget(TargetGraphNode):
    """@brief Target base class for CPU cores."""

    @property
    def name(self) -> str:
        """@brief CPU type name."""
        raise NotImplementedError()

    @property
    def core_number(self) -> int:
        raise NotImplementedError()

    @property
    def elf(self) -> Optional["ELFBinaryFile"]:
        raise NotImplementedError()

    @elf.setter
    def elf(self, filename: "ELFBinaryFile") -> None:
        raise NotImplementedError()

    def set_reset_catch(self, reset_type: Target.ResetType) -> None:
        raise NotImplementedError()

    def clear_reset_catch(self, reset_type: Target.ResetType) -> None:
        raise NotImplementedError()

    def set_target_context(self, context: "DebugContext") -> None:
        raise NotImplementedError()

    def exception_number_to_name(self, exc_num: int) -> Optional[str]:
        raise NotImplementedError()



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import string
import io
from typing import (IO, Optional, Sequence)

from . import conversion

## ASCII printable characters not including whitespace that changes line position.
_PRINTABLE = string.digits + string.ascii_letters + string.punctuation + ' '

def format_hex_width(value: int, width: int) -> str:
    """@brief Formats the value as hex of the specified bit width.

    @param value Integer value to be formatted.
    @param width Bit width, must be one of 8, 16, 32, 64.
    @return String with (width / 8) hex digits. Does not have a "0x" prefix.
    """
    if width == 8:
        return "%02x" % value
    elif width == 16:
        return "%04x" % value
    elif width == 32:
        return "%08x" % value
    elif width == 64:
        return "%016x" % value
    else:
        raise ValueError("unrecognized register width (%d)" % width)

def dump_hex_data(
        data: Sequence[int],
        start_address: int = 0,
        width: int = 8,
        output: Optional[IO[str]] = None,
        print_ascii: bool = True
    ) -> None:
    """@brief Prints a canonical hex dump of the given data.

    Each line of the output consists of an address column, the data as hex, and a printable ASCII
    representation of the data.

    The @a width parameter controls grouping of the hex bytes in the output. The bytes of the
    provided data are progressively read as little endian values of the specified bit width, then
    printed at that width. For example, for input data of [0x61 0x62 0x63 0x64], if @width is set to
    8 the output will be "61 62 63 64", for 16 it will be printed as "6261 6463", and for 32 bit
    width it will be shown as "64636261". A space is inserted after each bit-width value, with an
    extra space every 4 bytes for 8 bit width.

    The output looks similar to this (width of 8):
    ```
    00000000:  85 89 70 0f  20 b1 ff bc  a9 0c c8 3c  bc a6 47 dd    ..p. ......<..G.
    00000010:  c8 c9 66 ab  59 c8 35 6c  57 94 00 c8  17 35 85 b2    ..f.Y.5lW....5..
    ```

    The output is always terminated with a newline.

    If you want a string instead of output to a file, use the dump_hex_data_to_str() function.

    @param data The data to print as hex. Can be a `bytes`, `bytearray`, or list of integers.
    @param start_address Address of the first byte of the data. Defaults to 0. If set to None,
        then the address column is not printed.
    @param width Controls grouping of the hex bytes in the output as described above. Must be one of
        (8, 16, 32, 64).
    @param output Optional file where the output will be written. If not provided, sys.stdout is
        used.
    @param print_ascii Whether to include the printable ASCII column. Defaults to True.
    """
    if output is None:
        output = sys.stdout
    if width == 8:
        line_width = 16
    elif width == 16:
        line_width = 8
    elif width == 32:
        line_width = 4
    elif width == 64:
        line_width = 2
    else:
        raise RuntimeError(f"unsupported width of {width}")

    def line_width_in_chars(elements: int) -> int:
        return elements * ((2 * width // 8) + 1)

    max_line_width = line_width_in_chars(line_width)
    i = 0
    while i < len(data):
        if start_address is not None:
            output.write("%08x:  " % (start_address + (i * (width // 8))))

        start_i = i
        while i < len(data):
            d = data[i]
            i += 1
            if width == 8:
                output.write("%02x " % d)
                if (i % 4 == 0) and not (i % line_width == 0):
                    output.write(" ")
            elif width == 16:
                output.write("%04x " % d)
            elif width == 32:
                output.write("%08x " % d)
            elif width == 64:
                output.write("%016x " % d)
            if i % line_width == 0:
                break
        actual_line_width = line_width_in_chars(i - start_i)

        if print_ascii:
            s = "|"
            for n in range(start_i, start_i + line_width):
                if n >= len(data):
                    break
                d = data[n]
                if width == 8:
                    d = [d]
                else:
                    d = list(conversion.nbit_le_list_to_byte_list([d], width))
                    d.reverse()
                s += "".join((chr(b) if (chr(b) in _PRINTABLE) else '.') for b in d)
            output.write(" " * (max_line_width - actual_line_width) + "   " + s + "|")

        output.write("\n")

def dump_hex_data_to_str(data, **kwargs):
    """@brief Returns a string with data formatted as hex.
    @see dump_hex_data()
    """
    sio = io.StringIO()
    dump_hex_data(data, output=sio, **kwargs)
    return sio.getvalue()



#!/usr/bin/env python
# pyOCD debugger
# Copyright (c) 2006-2018 Arm Limited
# Copyright (c) 2020 Cypress Semiconductor Corporation
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import logging
import argparse
import json

from .. import __version__
from .. import target
from ..core.session import Session
from ..core.helpers import ConnectHelper
from ..gdbserver import GDBServer
from ..utility.cmdline import (split_command_line, convert_session_options)
from ..probe.pydapaccess import DAPAccess
from ..core.session import Session
from ..coresight.generic_mem_ap import GenericMemAPTarget

LOG = logging.getLogger(__name__)

LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}

SUPPORTED_TARGETS = list(sorted(target.TARGET.keys()))
DEBUG_LEVELS = list(LEVELS.keys())

class GDBServerTool(object):
    def __init__(self):
        self.args = None
        self.gdb_server_settings = None
        self.echo_msg = None

    def build_parser(self):
        # Build epilog with list of targets.
        epilog = "Available targets for use with --target option: " + ", ".join(SUPPORTED_TARGETS)

        # Keep args in snyc with flash_tool.py when possible
        parser = argparse.ArgumentParser(description='PyOCD GDB Server', epilog=epilog)
        parser.add_argument('--version', action='version', version=__version__)
        parser.add_argument('--config', metavar="PATH", default=None, help="Use a YAML config file.")
        parser.add_argument("--no-config", action="store_true", default=None, help="Do not use a configuration file.")
        parser.add_argument("--pack", metavar="PATH", help="Path to a CMSIS Device Family Pack")
        parser.add_argument("-p", "--port", dest="port_number", type=int, default=3333, help="Set the port number that GDB server will open (default 3333).")
        parser.add_argument("-sc", "--semihost-console", dest="semihost_console_type", default=None, choices=('telnet', 'stdx'), help="Console for semihosting.")
        parser.add_argument("-T", "--telnet-port", dest="telnet_port", type=int, default=4444, help="Specify the telnet port for semihosting (default 4444).")
        parser.add_argument("--allow-remote", dest="serve_local_only", default=True, action="store_false", help="Allow remote TCP/IP connections (default is no).")
        parser.add_argument("-b", "--board", dest="board_id", default=None, help="Connect to board by board ID. Use -l to list all connected boards. Only a unique part of the board ID needs to be provided.")
        parser.add_argument("-l", "--list", action="store_true", dest="list_all", default=False, help="List all connected boards.")
        parser.add_argument("--list-targets", action="store_true", dest="list_targets", default=False, help="List all available targets.")
        parser.add_argument("--json", action="store_true", dest="output_json", default=False, help="Output lists in JSON format. Only applies to --list and --list-targets.")
        parser.add_argument("-d", "--debug", dest="debug_level", choices=DEBUG_LEVELS, default='info', help="Set the level of system logging output. Supported choices are: " + ", ".join(DEBUG_LEVELS), metavar="LEVEL")
        parser.add_argument("-t", "--target", dest="target_override", default=None, help="Override target to debug.", metavar="TARGET")
        parser.add_argument("-n", "--nobreak", dest="no_break_at_hardfault", action="store_true", help="Disable halt at hardfault handler. (Deprecated)")
        parser.add_argument("-r", "--reset-break", dest="break_on_reset", action="store_true", help="Halt the target when reset. (Deprecated)")
        parser.add_argument("-C", "--vector-catch", default='h', help="Enable vector catch sources, one letter per enabled source in any order, or 'all' or 'none'. (h=hard fault, b=bus fault, m=mem fault, i=irq err, s=state err, c=check err, p=nocp, r=reset, a=all, n=none). (Default is hard fault.)")
        parser.add_argument("-s", "--step-int", dest="step_into_interrupt", default=None, action="store_true", help="Allow single stepping to step into interrupts.")
        parser.add_argument("-f", "--frequency", dest="frequency", default=None, type=int, help="Set the SWD clock frequency in Hz.")
        parser.add_argument("-o", "--persist", dest="persist", default=None, action="store_true", help="Keep GDB server running even after remote has detached.")
        parser.add_argument("-bh", "--soft-bkpt-as-hard", dest="soft_bkpt_as_hard", default=False, action="store_true", help="Replace software breakpoints with hardware breakpoints (ignored).")
        group = parser.add_mutually_exclusive_group()
        group.add_argument("-ce", "--chip_erase", action="store_true", help="Use chip erase when programming.")
        group.add_argument("-se", "--sector_erase", action="store_true", help="Use sector erase when programming.")
        # -Currently "--unlock" does nothing since kinetis parts will automatically get unlocked
        parser.add_argument("-u", "--unlock", action="store_true", default=False, help="Unlock the device.")
        # reserved: "-a", "--address"
        # reserved: "-s", "--skip"
        parser.add_argument("-hp", "--hide_progress", action="store_true", default=None, help="Don't display programming progress.")
        parser.add_argument("-fp", "--fast_program", action="store_true", default=None, help="Use only the CRC of each page to determine if it already has the same data.")
        parser.add_argument("-S", "--semihosting", dest="enable_semihosting", action="store_true", default=None, help="Enable semihosting.")
        parser.add_argument("-G", "--gdb-syscall", dest="semihost_use_syscalls", action="store_true", default=None, help="Use GDB syscalls for semihosting file I/O.")
        parser.add_argument("-c", "--command", dest="commands", metavar="CMD", action='append', nargs='+', help="Run command (OpenOCD compatibility).")
        parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
        parser.add_argument("--elf", metavar="PATH", help="Optionally specify ELF file being debugged.")
        parser.add_argument("-O", "--option", metavar="OPTION", action="append", help="Set session option of form 'OPTION=VALUE'.")
        parser.add_argument("--no-deprecation-warning", action="store_true", help="Do not warn about pyocd-gdbserver being deprecated.")
        self.parser = parser
        return parser

    def get_chip_erase(self, args):
        # Determine programming mode
        chip_erase = "auto"
        if args.chip_erase:
            chip_erase = "chip"
        elif args.sector_erase:
            chip_erase = "sector"
        return chip_erase

    def get_vector_catch(self, args):
        vector_catch = args.vector_catch.lower()

        # Handle deprecated options.
        if args.break_on_reset:
            vector_catch += 'r'
        if args.no_break_at_hardfault:
            # Must handle all case specially since we can't just filter 'h'.
            if vector_catch == 'all' or 'a' in vector_catch:
                vector_catch = 'bmiscpr' # Does not include 'h'.
            else:
                vector_catch = vector_catch.replace('h', '')

        return vector_catch

    def get_gdb_server_settings(self, args):
        # Set gdb server settings
        return {
            'gdbserver_port' : self.args.port_number,
            'step_into_interrupt' : args.step_into_interrupt,
            'persist' : args.persist,
            'chip_erase': self.get_chip_erase(args),
            'hide_programming_progress' : args.hide_progress,
            'fast_program' : args.fast_program,
            'enable_semihosting' : args.enable_semihosting,
            'semihost_console_type' : args.semihost_console_type,
            'telnet_port' : args.telnet_port,
            'semihost_use_syscalls' : args.semihost_use_syscalls,
            'serve_local_only' : args.serve_local_only,
            'vector_catch' : self.get_vector_catch(args),
        }


    def setup_logging(self, args):
        format = "%(relativeCreated)07d:%(levelname)s:%(module)s:%(message)s"
        level = LEVELS.get(args.debug_level, logging.NOTSET)
        logging.basicConfig(level=level, format=format)

    def process_commands(self, commands):
        """@brief Handle OpenOCD commands for compatibility."""
        if commands is None:
            return
        for cmd_list in commands:
            try:
                cmd_list = split_command_line(cmd_list)
                cmd = cmd_list[0]
                if cmd == 'gdb_port':
                    if len(cmd_list) < 2:
                        print("Missing port argument")
                    else:
                        self.args.port_number = int(cmd_list[1], base=0)
                elif cmd == 'telnet_port':
                    if len(cmd_list) < 2:
                        print("Missing port argument")
                    else:
                        self.gdb_server_settings['telnet_port'] = int(cmd_list[1], base=0)
                elif cmd == 'echo':
                    self.echo_msg = ' '.join(cmd_list[1:])
                else:
                    print("Unsupported command: %s" % ' '.join(cmd_list))
            except IndexError:
                pass

    def server_listening(self, note):
        if self.echo_msg is not None:
            print(self.echo_msg, file=sys.stderr)
            sys.stderr.flush()

    def disable_logging(self):
        logging.getLogger().setLevel(logging.FATAL)

    def list_boards(self):
        self.disable_logging()

        if not self.args.output_json:
            ConnectHelper.list_connected_probes()
        else:
            status = 0
            error = ""
            try:
                all_mbeds = ConnectHelper.get_sessions_for_all_connected_probes(blocking=False)
            except Exception as e:
                all_mbeds = []
                status = 1
                error = str(e)
                if not self.args.output_json:
                    raise

            boards = []
            obj = {
                'pyocd_version' : __version__,
                'version' : { 'major' : 1, 'minor' : 0 },
                'status' : status,
                'boards' : boards,
                }

            if status != 0:
                obj['error'] = error

            for mbed in all_mbeds:
                d = {
                    'unique_id' : mbed.probe.unique_id,
                    'info' : mbed.board.description,
                    'board_name' : mbed.board.name,
                    'target' : mbed.board.target_type,
                    'vendor_name' : mbed.probe.vendor_name,
                    'product_name' : mbed.probe.product_name,
                    }
                boards.append(d)

            print(json.dumps(obj, indent=4))

    def list_targets(self):
        self.disable_logging()

        if self.args.output_json:
            targets = []
            obj = {
                'pyocd_version' : __version__,
                'version' : { 'major' : 1, 'minor' : 0 },
                'status' : 0,
                'targets' : targets
                }

            for name in SUPPORTED_TARGETS:
                s = Session(None) # Create empty session
                t = target.TARGET[name](s)
                d = {
                    'name' : name,
                    'part_number' : t.part_number,
                    }
                if t._svd_location is not None:
                    svdPath = t._svd_location.filename
                    if os.path.exists(svdPath):
                        d['svd_path'] = svdPath
                targets.append(d)

            print(json.dumps(obj, indent=4))
        else:
            for t in SUPPORTED_TARGETS:
                print(t)

    def run(self, args=None):
        self.args = self.build_parser().parse_args(args)
        self.gdb_server_settings = self.get_gdb_server_settings(self.args)
        self.setup_logging(self.args)
        DAPAccess.set_args(self.args.daparg)

        if not self.args.no_deprecation_warning:
            LOG.warning("pyocd-gdbserver is deprecated; please use the new combined pyocd tool.")

        self.process_commands(self.args.commands)

        gdb = None
        gdbs = []
        if self.args.list_all == True:
            self.list_boards()
        elif self.args.list_targets == True:
            self.list_targets()
        else:
            try:
                # Build dict of session options.
                sessionOptions = convert_session_options(self.args.option)
                sessionOptions.update(self.gdb_server_settings)

                session = ConnectHelper.session_with_chosen_probe(
                    config_file=self.args.config,
                    no_config=self.args.no_config,
                    pack=self.args.pack,
                    unique_id=self.args.board_id,
                    target_override=self.args.target_override,
                    frequency=self.args.frequency,
                    **sessionOptions)
                if session is None:
                    print("No board selected")
                    return 1
                with session:
                    # Set ELF if provided.
                    if self.args.elf:
                        session.board.target.elf = self.args.elf
                    for core_number, core in session.board.target.cores.items():
                        if isinstance(session.board.target.cores[core_number], GenericMemAPTarget):
                            continue

                        gdb = GDBServer(session, core=core_number)
                        # Only subscribe to the server for the first core, so echo messages aren't printed
                        # multiple times.
                        if not gdbs:
                            session.subscribe(self.server_listening, GDBServer.GDBSERVER_START_LISTENING_EVENT, gdb)
                        session.gdbservers[core_number] = gdb
                        gdbs.append(gdb)
                        gdb.start()
                    gdb = gdbs[0]
                    while gdb.is_alive():
                        gdb.join(timeout=0.5)
            except KeyboardInterrupt:
                for gdb in gdbs:
                    gdb.stop()
            except Exception as e:
                LOG.error("uncaught exception: %s" % e, exc_info=Session.get_current().log_tracebacks)
                for gdb in gdbs:
                    gdb.stop()
                return 1

        # Successful exit.
        return 0

def main():
    sys.exit(GDBServerTool().run())

if __name__ == '__main__':
    main()



# pyOCD debugger
# Copyright (c) 2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

LOG = logging.getLogger(__name__)

## Whether the warning about no libusb was printed already.
#
# Used to prevent warning spewage if repeatedly scanning for probes, such as when ConnectHelper
# is used in blocking mode and no probes are connected.
did_show_no_libusb_warning = False

## Set of VID/PID tuples for which libusb errors have been reported.
#
# Used to prevent spewing lots of errors for the same devices when repeatedly scanning for probes.
libusb_error_device_set = set()

def show_no_libusb_warning():
    """@brief Logs a warning about missing libusb library only the first time it is called."""
    global did_show_no_libusb_warning
    if not did_show_no_libusb_warning:
        LOG.warning("STLink, CMSIS-DAPv2 and PicoProbe probes are not supported because no libusb library was found.")
        did_show_no_libusb_warning = True

def should_show_libusb_device_error(vidpid):
    """@brief Returns whether a debug warning should be shown for the given VID/PID pair.

    The first time a given VID/PID is passed to this function, the result will be True. Any
    subsequent times, False will be returned for the same VID/PID pair.

    @param vidpi A bi-tuple of USB VID and PID, in that order.
    """
    should_log = vidpid not in libusb_error_device_set
    libusb_error_device_set.add(vidpid)
    return should_log



# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2020 Patrick Huesmann
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from time import sleep
from typing import (Optional, TextIO, TYPE_CHECKING)

from .sink import TraceEventSink
from .events import (TraceEvent, TraceITMEvent)
from .swo import SWOParser
from ..coresight.itm import ITM
from ..coresight.tpiu import TPIU
from ..core.target import Target
from ..core import exceptions
from ..probe.debug_probe import DebugProbe
from ..utility.server import StreamServer

if TYPE_CHECKING:
    from ..core.session import Session
    from ..utility.notification import Notification

LOG = logging.getLogger(__name__)

class SWVEventSink(TraceEventSink):
    """@brief Trace event sink that converts ITM packets to a text stream."""

    def __init__(self, console: TextIO) -> None:
        """@brief Constructor.
        @param self
        @param console File-like object to which SWV data will be written.
        """
        self._console = console

    def receive(self, event: TraceEvent) -> None:
        """@brief Handle an SWV trace event.
        @param self
        @param event An instance of TraceITMEvent. If the event is not this class, or isn't
            for ITM port 0, then it will be ignored. The individual bytes of 16- or 32-bit events
            will be extracted and written to the console.
        """
        if not isinstance(event, TraceITMEvent):
            return

        # Extract bytes.
        if event.width == 1:
            data = chr(event.data)
        elif event.width == 2:
            data = chr(event.data & 0xff) + chr((event.data >> 8) & 0xff)
        elif event.width == 4:
            data = (chr(event.data & 0xff)
                    + chr((event.data >> 8) & 0xff)
                    + chr((event.data >> 16) & 0xff)
                    + chr((event.data >> 24) & 0xff))
        else:
            return

        self._console.write(data)

class SWVReader(threading.Thread):
    """@brief Sets up SWV and processes data in a background thread."""

    def __init__(self, session: "Session", core_number: int = 0, lock: Optional[threading.Lock] = None) -> None:
        """@brief Constructor.
        @param self
        @param session The Session instance.
        @param core_number The number of the core being traced. Default is core 0.
        """
        super().__init__(name="SWVReader", daemon=True)
        self._session = session
        self._core_number = core_number
        self._shutdown_event = threading.Event()
        self._swo_clock = 0
        self._lock = lock

        target = self._session.target
        assert target
        self._target = target
        self._core = target.cores[core_number]

        self._session.subscribe(self._reset_handler, Target.Event.POST_RESET, self._core)

    def init(self, sys_clock: int, swo_clock: int, console: TextIO) -> bool:
        """@brief Configures trace graph and starts thread.

        This method performs all steps required to start up SWV. It first calls the target's
        trace_start() method, which allows for target-specific trace initialization. Then it
        configures the TPIU and ITM modules. A simple trace data processing graph is created that
        connects an SWVEventSink with a SWOParser. Finally, the reader thread is started.

        If the debug probe or target do not support SWO, a warning is printed and False returns,
        but nothing else is done (no exception raised).

        @param self
        @param sys_clock System clock frequency in Hertz, from which the SWO clock is derived.
        @param swo_clock Desired SWO output frequency in Hertz.
        @param console File-like object to which SWV data will be written.

        @return Boolean indicating whether the SWV reader was successfully started.
        """
        self._swo_clock = swo_clock

        assert self._session.probe
        if DebugProbe.Capability.SWO not in self._session.probe.capabilities:
            LOG.warning(f"SWV not initalized: Probe {self._session.probe.unique_id} does not support SWO")
            return False

        itm = self._target.get_first_child_of_type(ITM)
        if not itm:
            LOG.warning("SWV not initalized: Target does not have ITM component")
            return False
        tpiu = self._target.get_first_child_of_type(TPIU)
        if not tpiu:
            LOG.warning("SWV not initalized: Target does not have TPIU component")
            return False

        self._target.trace_start()

        itm.init()
        itm.enable()
        tpiu.init()

        if tpiu.set_swo_clock(swo_clock, sys_clock):
            LOG.info("Set SWO clock to %d", swo_clock)
        else:
            LOG.warning("SWV not initalized: Failed to set SWO clock rate")
            return False

        self._parser = SWOParser(self._core)
        self._sink = SWVEventSink(console)
        self._parser.connect(self._sink)

        self.start()

        return True

    def stop(self) -> None:
        """@brief Stops processing SWV data.

        The reader thread is terminated first, then the ITM is disabled. The last step is to call
        the target's trace_stop() method.

        Does nothing if the init() method did not complete successfully.
        """
        if not self.is_alive():
            return

        self._shutdown_event.set()
        self.join()

        # init() should never have started the SWV thread unless the target has ITM and TPIU.
        itm = self._target.get_first_child_of_type(ITM)
        assert itm
        itm.disable()

        self._target.trace_stop()

    def run(self) -> None:
        """@brief SWV reader thread routine.

        Starts the probe receiving SWO data by calling DebugProbe.swo_start(). For as long as the
        thread runs, it reads SWO data from the probe and passes it to the SWO parser created in
        init(). When the thread is signaled to stop, it calls DebugProbe.swo_stop() before exiting.
        """
        assert self._session.probe

        if self._lock:
            self._lock.acquire()

        swv_raw_server = StreamServer(
                            self._session.options.get('swv_raw_port'),
                            serve_local_only=self._session.options.get('serve_local_only'),
                            name="SWV raw",
                            is_read_only=True) \
                         if self._session.options.get('swv_raw_enable') else None

        # Stop SWO first in case the probe already had it started. Ignore if this fails.
        try:
            self._session.probe.swo_stop()
        except exceptions.ProbeError:
            pass
        self._session.probe.swo_start(self._swo_clock)

        while not self._shutdown_event.is_set():
            data = self._session.probe.swo_read()
            if data:
                if swv_raw_server:
                    swv_raw_server.write(data)
                self._parser.parse(data)

            if self._lock:
                self._lock.release()

            sleep(0.001)

            if self._lock:
                self._lock.acquire()

        self._session.probe.swo_stop()

        if swv_raw_server:
            swv_raw_server.stop()

        if self._lock:
            self._lock.release()

    def _reset_handler(self, notification: "Notification") -> None:
        """@brief Reset notification handler.

        If the target is reset while the SWV reader is running, then the Target::trace_start()
        method is called to reinit trace output.
        """
        if self.is_alive():
            self._target.trace_start()




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Any, TYPE_CHECKING)

from ..core import exceptions

if TYPE_CHECKING:
    from ..core.target import Target
    from types import TracebackType

class Autoflush:
    """@brief Context manager for performing flushes.

    Pass a Target instance to the constructor, and when the context exits, the target will be
    automatically flushed. If a TransferError or subclass, such as TransferFaultError, is raised
    within the context, then the flush will be skipped.

    The parameter passed to the constructor can actually be any object with a `flush()` method,
    due to Python's dynamic dispatch.
    """

    def __init__(self, target: "Target") -> None:
        """@brief Constructor.

        @param self The object.
        @param target Object on which the flush will be performed. Normally this is a Target
            instance.
        """
        self._target = target

    def __enter__(self) -> "Autoflush":
        return self

    def __exit__(self, exc_type: type, value: Any, traceback: "TracebackType") -> bool:
        if exc_type is None or not issubclass(exc_type, exceptions.TransferError):
            self._target.flush()
        return False
    



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import exceptions
from ..core.plugin import load_plugin_classes_of_type
from .debug_probe import DebugProbe

## @brief Dictionary of loaded probe plugins indexed by name.
PROBE_CLASSES = {}

class DebugProbeAggregator(object):
    """@brief Simple class to enable collecting probes of all supported probe types."""

    @staticmethod
    def _get_probe_classes(unique_id):
        """@brief Return probe classes to query based on the unique ID string."""
        probe_type = None
        if unique_id is not None:
            fields = unique_id.split(':', 1)
            if len(fields) > 1:
                probe_type = fields[0].lower()
                unique_id = fields[1]

        if probe_type is None:
            klasses = PROBE_CLASSES.values()
        else:
            # Perform a case-insensitive match.
            klasses = [PROBE_CLASSES[k] for k in PROBE_CLASSES if k.lower() == probe_type]
            if not klasses:
                raise exceptions.Error("unknown debug probe type '{}'".format(probe_type))

        return klasses, unique_id, (probe_type is not None)

    @staticmethod
    def get_all_connected_probes(unique_id=None):
        klasses, unique_id, is_explicit = DebugProbeAggregator._get_probe_classes(unique_id)

        probes = []

        # First look for a match against the full ID, as this can be more efficient for certain probes.
        if unique_id is not None:
            for cls in klasses:
                probe = cls.get_probe_with_id(unique_id, is_explicit)
                if probe is not None:
                    return [probe]

        # No full match, so ask probe classes for probes.
        for cls in klasses:
            probes += cls.get_all_connected_probes(unique_id, is_explicit)

        # Filter by unique ID.
        if unique_id is not None:
            unique_id = unique_id.lower()
            probes = [probe for probe in probes if (unique_id in probe.unique_id.lower())]

        return probes

    @classmethod
    def get_probe_with_id(cls, unique_id):
        klasses, unique_id, is_explicit = DebugProbeAggregator._get_probe_classes(unique_id)

        for cls in klasses:
            probe = cls.get_probe_with_id(unique_id, is_explicit)
            if probe is not None:
                return probe
        return None

# Load plugins when this module is loaded.
load_plugin_classes_of_type('pyocd.probe', PROBE_CLASSES, DebugProbe)



# pyOCD debugger
# Copyright (c) 2022-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from pathlib import PurePath
from typing import (Any, Dict, cast, TYPE_CHECKING)

from ...core.memory_map import (
    FlashRegion,
    MemoryMap,
    MemoryRange,
    MemoryType,
    RamRegion,
)
from .flash_algo import (PackFlashAlgo, FlashAlgoException)

if TYPE_CHECKING:
    from ...coresight.coresight_target import CoreSightTarget
    from ...core.memory_map import (MemoryMap)

LOG = logging.getLogger(__name__)

class FlmFlashRegionBuilder:
    """
    @brief Finishes the process of constructing flash regions with algos based on FLM files.

    If a flash region passed to finalise_region() is missing an algo dict but has an associated FLM
    algo file, the FLM will be loaded.
    """

    def __init__(self, target: "CoreSightTarget", memory_map: "MemoryMap") -> None:
        self._target = target
        self._session = target.session
        self._memory_map = memory_map

    def finalise_region(self, region: FlashRegion) -> bool:
        """@brief Load FLM file for the given flash region.

        The region is ignored if it has an algo dict already, or if it doesn't have an FLM file. Otherwise
        the algo dict is constructed from the FLM and subregions are added for each of the algo's sector
        size ranges.

        @return Boolean indicating whether the region was finalised successfully.
        """
        try:
            # If the region doesn't have an algo dict but does have an FLM file, try to load
            # the FLM and create the algo dict.
            if (region.algo is None) and (region.flm is not None):
                if isinstance(region.flm, (str, PurePath)):
                    flm_path = self._session.find_user_file(None, [str(region.flm)])
                    if flm_path is not None:
                        LOG.info("Creating flash algo for region %s from: %s", region.name, flm_path)
                        pack_algo = PackFlashAlgo(flm_path)
                    else:
                        LOG.warning("Failed to find FLM file: %s", region.flm)
                        return False
                elif isinstance(region.flm, PackFlashAlgo):
                    pack_algo = region.flm
                else:
                    LOG.warning("Flash region %s flm attribute is unexpected type", region)
                    return False

                # Log details of this flash algo if the debug option is enabled.
                if self._session.options.get("debug.log_flm_info"):
                    LOG.debug("Flash algo info: %s", pack_algo.flash_info)

                # Get the page size. If it's unreasonably small, then use the smallest sector size.
                page_size = pack_algo.page_size
                if page_size <= 32:
                    page_size = min(s[1] for s in pack_algo.sector_sizes)

                # Select the RAM to use for the algo.
                try:
                    ram_for_algo = self._select_flash_ram(region)
                except RuntimeError:
                    return False

                # Create the algo dict from the FLM.
                algo = pack_algo.get_pyocd_flash_algo(page_size, ram_for_algo)

                # If we got a valid algo from the FLM, set it on the region.
                if algo is not None:
                    region.algo = algo

                # Set region page/sector sizes and algorithm range; add sector subregions.
                self._update_flash_attributes(region, pack_algo, page_size, algo)

            return True
        except FlashAlgoException as algo_err:
            LOG.warning("Failed to load flash algorithm for region '%s' (%x-%x): %s",
                    region.name, region.start, region.end, algo_err)
            return False

    def _select_flash_ram(self, region: FlashRegion) -> RamRegion:
        """@brief Choose the RAM region to use for the given flash region's algo.

        @exception RuntimeError No RAM region is available.
        """
        # See if an explicit RAM range was specified for the algo.
        if hasattr(region, '_RAMstart'):
            ram_start = region._RAMstart

            # The region size comes either from the RAMsize attribute, the containing region's
            # bounds, or a large, arbitrary value.
            if hasattr(region, '_RAMsize'):
                ram_size = region._RAMsize
            else:
                containing_region = self._memory_map.get_region_for_address(ram_start)
                if containing_region is not None:
                    ram_size = containing_region.length - (ram_start - containing_region.start)
                else:
                    # No size specified, and the RAMstart attribute is outside of a known region,
                    # so just use a mid-range arbitrary size.
                    ram_size = 16 * 1024

            ram_for_algo = RamRegion(start=ram_start, length=ram_size)
        else:
            # No RAM addresses were given, so go with the RAM marked default.
            ram_for_algo = cast(RamRegion, self._memory_map.get_default_region_of_type(MemoryType.RAM))
            # Must have a default ram.
            if ram_for_algo is None:
                LOG.warning(f"CMSIS-Pack device {self._target.part_number} has no default RAM defined; cannot program flash")
                raise RuntimeError("no default RAM")

        return ram_for_algo

    def _update_flash_attributes(
            self,
            region: FlashRegion,
            pack_algo: PackFlashAlgo,
            page_size: int,
            algo: Dict[str, Any],
            ) -> None:
        """Depending on the sector size(s) defined by the flash algorithm, either simply set
        the parent flash region's attributes or create sector size subregions."""
        # First set the region's start and end if they weren't set.
        if region.start == region.end:
            # Directly access the attributes. Normally region start/end are not settable to
            # prevent modification of regions in a memory map such that they overlap.
            region._start = pack_algo.flash_start
            region._end = pack_algo.flash_start + pack_algo.flash_size - 1

        # Don't need to create subregions if there is a single sector size and its range
        # starts at the same address and is equal or larger than the parent flash region.
        sector_sizes = list(pack_algo.iter_sector_size_ranges())
        create_subregions = not (len(sector_sizes) == 1
                                and sector_sizes[0][0].start == region.start
                                and sector_sizes[0][0].end >= region.end)

        if create_subregions:
            self._add_flash_subregions(region, pack_algo, page_size, algo)
        else:
            # Set attributes on parent flash region. The parent region still has to have these attributes
            # even though there are subregions.
            region.attributes['page_size'] = page_size
            region.attributes['sector_size'] = sector_sizes[0][1]

    def _add_flash_subregions(
            self,
            region: FlashRegion,
            pack_algo: PackFlashAlgo,
            page_size: int,
            algo: Dict[str, Any],
            ) -> None:
        """@brief Create subregions of the parent flash region for each sector size.

        The overall range of combined sector sizes doesn't necessarily fill the parent region's
        entire size. Conversely, the algorithm may define more sectors than fit in the parent
        region.
        """
        max_sector_size = 0

        # Create subregions.
        for range, sector_size in pack_algo.iter_sector_size_ranges():
            # Limit subregion range to parent region size. There are cases (eg nRF5340, nRF9160) where
            # the flash algo defines a larger flash memory, then the DFP's <memory> attribute sets a
            # smaller value.
            if not region.contains_range(range):
                range = MemoryRange(max(region.start, range.start), end=min(region.end, range.end))
                if range.is_empty:
                    continue

            # Track maximum sector size.
            max_sector_size = max(max_sector_size, sector_size)

            # Limit page size.
            if page_size > sector_size:
                region_page_size = sector_size
                LOG.warning(f"Page size ({page_size}) is larger than sector size ({sector_size}) for flash "
                            f"region {region.name}; using sector size")
            else:
                region_page_size = page_size

            # Construct unique region name.
            region_name = region.name + f"_{sector_size:#x}"

            subregion = FlashRegion(
                name=region_name,
                access=region.access,
                start=range.start,
                end=range.end,
                sector_size=sector_size,
                page_size=region_page_size,
                erased_byte_value=pack_algo.flash_info.value_empty,
                algo=algo,
            )

            region.submap.add_region(subregion)

        # Set attributes on parent flash region. The parent region still has to have these attributes
        # even though there are subregions.
        region.attributes['page_size'] = page_size
        region.attributes['sector_size'] = max_sector_size



# pyOCD debugger
# Copyright (c) 2006-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# Copyright (c) 2022 Harper Weigle
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import platform
import six
import threading
import queue
from typing import Optional

from .interface import Interface
from .common import (
    filter_device_by_usage_page,
    generate_device_unique_id,
    is_known_cmsis_dap_vid_pid,
    )
from ..dap_access_api import DAPAccessIntf
from ....utility.compatibility import to_str_safe
from ....utility.timeout import Timeout

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

try:
    import hid
except ImportError:
    IS_AVAILABLE = False
else:
    IS_AVAILABLE = True

# OS flags.
_IS_DARWIN = (platform.system() == 'Darwin')
_IS_WINDOWS = (platform.system() == 'Windows')

class HidApiUSB(Interface):
    """@brief CMSIS-DAP USB interface class using hidapi backend."""

    isAvailable = IS_AVAILABLE

    HIDAPI_MAX_PACKET_COUNT = 30

    def __init__(self, dev, info: dict):
        super().__init__()
        # Vendor page and usage_id = 2
        self.vid = info['vendor_id']
        self.pid = info['product_id']
        self.vendor_name = info['manufacturer_string'] or f"{self.vid:#06x}"
        self.product_name = info['product_string'] or f"{self.pid:#06x}"
        self.serial_number = info['serial_number'] \
                or generate_device_unique_id(self.vid, self.pid, six.ensure_str(info['path']))
        self.device_info = info
        self.device = dev
        self.closed = True
        self.thread = None
        self.read_sem = threading.Semaphore(0)
        self.closed_event = threading.Event()
        self.received_data: queue.SimpleQueue[bytes] = queue.SimpleQueue()
        self._read_thread_did_exit: bool = False
        self._read_thread_exception: Optional[Exception] = None

    def set_packet_count(self, count):
        # hidapi for macos has an arbitrary limit on the number of packets it will queue for reading.
        # Even though we have a read thread, it doesn't hurt to limit the packet count since the limit
        # is fairly high.
        if _IS_DARWIN:
            count = min(count, self.HIDAPI_MAX_PACKET_COUNT)
        self.packet_count = count

    def open(self):
        try:
            self.device.open_path(self.device_info['path'])
        except IOError as exc:
            raise DAPAccessIntf.DeviceError("Unable to open device: " + str(exc)) from exc

        # Windows does not use the receive thread because it causes packet corruption for some reason.
        if not _IS_WINDOWS:
            # Make certain the closed event is clear.
            self.closed_event.clear()

            # Start RX thread
            self.thread = threading.Thread(target=self.rx_task)
            self.thread.daemon = True
            self.thread.start()

        self.closed = False

    def rx_task(self):
        try:
            while not self.closed_event.is_set():
                self.read_sem.acquire()
                if not self.closed_event.is_set():
                    read_data = bytes(self.device.read(self.packet_size))

                    # This trace log is commented out to reduce clutter, but left in to leave available
                    # when debugging rx_task issues.
                    # if TRACE.isEnabledFor(logging.DEBUG):
                    #     # Strip off trailing zero bytes to reduce clutter.
                    #     TRACE.debug("  USB RD < (%d) %s", len(read_data),
                    #                 ' '.join([f'{i:02x}' for i in read_data.rstrip(b'\x00')]))

                    self.received_data.put(read_data)
        except Exception as err:
            TRACE.debug("rx_task exception: %s", err)
            self._read_thread_exception = err
        finally:
            self._swo_thread_did_exit = True

    @staticmethod
    def get_all_connected_interfaces():
        """@brief Returns all the connected devices with CMSIS-DAP in the name.

        returns an array of HidApiUSB (Interface) objects
        """

        devices = hid.enumerate()

        boards = []

        for deviceInfo in devices:
            product_name = to_str_safe(deviceInfo['product_string'])
            known_cmsis_dap = is_known_cmsis_dap_vid_pid(deviceInfo['vendor_id'], deviceInfo['product_id'])
            if ("CMSIS-DAP" not in product_name) and (not known_cmsis_dap):
                # Check the device path as a backup. Even though we can't get the interface name from
                # hidapi, it may appear in the path. At least, it does on macOS.
                device_path = to_str_safe(deviceInfo['path'])
                if "CMSIS-DAP" not in device_path:
                    # Skip non cmsis-dap devices
                    continue

            vid = deviceInfo['vendor_id']
            pid = deviceInfo['product_id']

            # Perform device-specific filtering.
            if filter_device_by_usage_page(vid, pid, deviceInfo['usage_page']):
                continue

            try:
                dev = hid.device(vendor_id=vid, product_id=pid, path=deviceInfo['path'])
            except IOError as exc:
                LOG.debug("Failed to open USB device: %s", exc)
                continue

            # Create the USB interface object for this device.
            new_board = HidApiUSB(dev, deviceInfo)
            boards.append(new_board)

        return boards

    def write(self, data):
        """@brief Write data on the OUT endpoint associated to the HID interface"""
        if TRACE.isEnabledFor(logging.DEBUG):
            TRACE.debug("  USB OUT> (%d) %s", len(data), ' '.join([f'{i:02x}' for i in data]))
        data.extend([0] * (self.packet_size - len(data)))
        if not _IS_WINDOWS:
            self.read_sem.release()
        self.device.write([0] + data)

    def read(self):
        """@brief Read data on the IN endpoint associated to the HID interface"""
        # Windows doesn't use the read thread, so read directly.
        if _IS_WINDOWS:
            read_data = bytes(self.device.read(self.packet_size))

            if TRACE.isEnabledFor(logging.DEBUG):
                # Strip off trailing zero bytes to reduce clutter.
                TRACE.debug("  USB IN < (%d) %s", len(read_data),
                            ' '.join([f'{i:02x}' for i in read_data.rstrip(b'\x00')]))

            return read_data

        # Check for terminated read thread.
        if self.closed:
            return b''
        elif self._read_thread_did_exit:
            raise DAPAccessIntf.DeviceError("Probe %s read thread exited unexpectedly" % self.serial_number) from self._read_thread_exception

        try:
            read_data = self.received_data.get(True, self.DEFAULT_USB_TIMEOUT_S)
        except queue.Empty:
            raise DAPAccessIntf.DeviceError(f"Timeout reading from probe {self.serial_number}") from None

        # Trace when the higher layer actually gets a packet previously read.
        if TRACE.isEnabledFor(logging.DEBUG):
            # Strip off trailing zero bytes to reduce clutter.
            TRACE.debug("  USB RD < (%d) %s", len(read_data),
                    ' '.join([f'{i:02x}' for i in read_data.rstrip(b'\x00')]))

        return read_data

    def close(self):
        """@brief Close the interface"""
        assert not self.closed_event.is_set()

        LOG.debug("closing interface")
        self.closed = True
        if not _IS_WINDOWS:
            self.closed_event.set()
            self.read_sem.release()
            assert self.thread
            self.thread.join()
            self.thread = None

            # Clear closed event, recreate read sem and receiveed data deque so they
            # are cleared and ready if we're re-opened.
            self.closed_event.clear()
            self.read_sem = threading.Semaphore(0)
            self.received_data = queue.SimpleQueue()
            self._read_thread_did_exit = False
            self._read_thread_exception = None
        self.device.close()



# pyOCD debugger
# Copyright (c) 2006-2013,2018 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Interface:

    @staticmethod
    def get_all_connected_interfaces():
        raise NotImplementedError()

    DEFAULT_USB_TIMEOUT_S = 10
    DEFAULT_USB_TIMEOUT_MS = DEFAULT_USB_TIMEOUT_S * 1000

    def __init__(self):
        self.vid = 0
        self.pid = 0
        self.vendor_name = ""
        self.product_name = ""
        self.serial_number = ""
        self.packet_count = 1
        self.packet_size = 64

    @property
    def has_swo_ep(self):
        return False

    @property
    def is_bulk(self):
        """@brief Whether the interface uses CMSIS-DAP v2 bulk endpoints."""
        return False

    def open(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def write(self, data):
        raise NotImplementedError()

    def read(self):
        raise NotImplementedError()

    def read_swo(self):
        raise NotImplementedError()

    def get_info(self):
        return self.vendor_name + " " + \
               self.product_name + " (" + \
               str(hex(self.vid)) + ", " + \
               str(hex(self.pid)) + ")"

    def get_packet_count(self):
        return self.packet_count

    def set_packet_count(self, count):
        # No interface level restrictions on count
        self.packet_count = count

    def set_packet_size(self, size):
        self.packet_size = size

    def get_packet_size(self):
        return self.packet_size

    def get_serial_number(self):
        return self.serial_number

    def __repr__(self):
        return f"<{type(self).__name__}@{id(self):x} {self.get_info()} {self.serial_number}>"



# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# Copyright (c) 2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from functools import reduce
from typing import (Any, Optional, Sequence, Tuple, Union)

def bitmask(*args: Union[int, Sequence[int], Tuple[int, int]]) -> int:
    """@brief Returns a mask with specified bit ranges set.

    An integer mask is generated based on the bits and bit ranges specified by the
    arguments. Any number of arguments can be provided. Each argument may be either
    a 2-tuple of integers, a list of integers, or an individual integer. The result
    is the combination of masks produced by the arguments.

    - 2-tuple: The tuple is a bit range with the first element being the MSB and the
          second element the LSB. All bits from LSB up to and included MSB are set.
    - list: Each bit position specified by the list elements is set.
    - int: The specified bit position is set.

    @return An integer mask value computed from the logical OR'ing of masks generated
      by each argument.

    Example:
    @code
      >>> hex(bitmask((23,17),1))
      0xfe0002
      >>> hex(bitmask([4,0,2],(31,24))
      0xff000015
    @endcode
    """
    mask = 0

    for a in args:
        if isinstance(a, tuple):
            hi, lo = a
            mask |= ((1 << (hi - lo + 1)) - 1) << lo
        elif isinstance(a, (list, set)):
            mask |= reduce(operator.or_, ((1 << b) for b in a))
        elif isinstance(a, int):
            mask |= 1 << a

    return mask

def bit_invert(value: int, width: int = 32) -> int:
    """@brief Return the bitwise inverted value of the argument given a specified width.

    @param value Integer value to be inverted.
    @param width Bit width of both the input and output. If not supplied, this defaults to 32.
    @return Integer of the bitwise inversion of @a value.
    """
    return ((1 << width) - 1) & (~value)

invert32 = bit_invert
"""@brief Return the 32-bit inverted value of the argument."""

def bfx(value: int, msb: int, lsb: int) -> int:
    """@brief Extract a value from a bitfield."""
    mask = bitmask((msb, lsb))
    return (value & mask) >> lsb

def bfxw(value: int, lsb: int, width: int) -> int:
    """@brief Extract a value from a bitfield given the LSb and width."""
    mask = bitmask((lsb + width, lsb))
    return (value & mask) >> lsb

def bfi(value: int, msb: int, lsb: int, field: int) -> int:
    """@brief Change a bitfield value."""
    mask = bitmask((msb, lsb))
    value &= ~mask
    value |= (field << lsb) & mask
    return value

class Bitfield:
    """@brief Represents a bitfield of a register."""

    def __init__(self, msb: int, lsb: Optional[int] = None, name: Optional[str] = None):
        self._msb = msb
        self._lsb = lsb if (lsb is not None) else msb
        self._name = name
        assert self._msb >= self._lsb

    @property
    def width(self) -> int:
        return self._msb - self._lsb + 1

    def get(self, value: int) -> int:
        """@brief Extract the bitfield value from a register value.
        @param self The Bitfield object.
        @param value Integer register value.
        @return Integer value of the bitfield extracted from `value`.
        """
        return bfx(value, self._msb, self._lsb)

    def set(self, register_value: int, field_value: int) -> int:
        """@brief Modified the bitfield in a register value.
        @param self The Bitfield object.
        @param register_value Integer register value.
        @param field_value New value for the bitfield. Must not be shifted into place already.
        @return Integer register value with the bitfield updated to `field_value`.
        """
        return bfi(register_value, self._msb, self._lsb, field_value)

    def __repr__(self) -> str:
        return "<{}@{:x} name={} {}:{}>".format(self.__class__.__name__, id(self), self._name, self._msb, self._lsb)

def msb(n: int) -> int:
    """@brief Return the bit number of the highest set bit."""
    ndx = 0
    while ( 1 < n ):
        n = ( n >> 1 )
        ndx += 1
    return ndx

def same(d1: Sequence[Any], d2: Sequence[Any]) -> bool:
    """@brief Test whether two sequences contain the same values.

    Unlike a simple equality comparison, this function works as expected when the two sequences
    are of different types, such as a list and bytearray. The sequences must return
    compatible types from indexing.
    """
    if len(d1) != len(d2):
        return False
    for i in range(len(d1)):
        if d1[i] != d2[i]:
            return False
    return True

def align_down(value: int, multiple: int) -> int:
    """@brief Return value aligned down to multiple."""
    return value // multiple * multiple

def align_up(value: int, multiple: int) -> int:
    """@brief Return value aligned up to multiple."""
    return (value + multiple - 1) // multiple * multiple

def round_up_div(value: int, divisor: int) -> int:
    """@brief Return value divided by the divisor, rounding up to the nearest multiple of the divisor."""
    return (value + divisor - 1) // divisor

def parity32_high(n: int) -> int:
    """@brief Compute parity over a 32-bit value.

    This function is intended to be used for computing parity over a 32-bit value transferred in an Arm
    ADI AP/DP register transfer. The result is returned in bit 32, ready to be OR'd into the register
    value to form the 33-bit data + parity for the AP/DP register transfer.

    @param n 32-bit integer.
    @return Integer with 1-bit parity placed at bit 32. The lower 32 bits are 0.
    """
    n ^= n >> 16
    n ^= n >> 8
    n ^= n >> 4
    n &= 0xf
    return (0xD32C0000 << n) & (1 << 32)

def twos_complement(value: int, width: int) -> int:
    """@brief Convert an unsigned int to signed.

    @param value Signed value as an unsigned int.
    @param width Bit width of _value_.
    @return Signed version of _value_.
    """
    signmask = 1 << (width - 1)
    if (value & signmask) == 0:
        # Mask off sign bit.
        return value & (signmask - 1)
    else:
        # Two's complement.
        return -bit_invert(value, width - 1) - 1



# pyOCD debugger
# Copyright (c) 2018-2020,2022 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from time import sleep
from typing import (Any, Callable, Dict, List, Optional, Sequence, Union, TYPE_CHECKING)

from .debug_probe import DebugProbe
from ..core.memory_interface import MemoryInterface
from ..core.plugin import Plugin
from ..core.options import OptionInfo
from ..coresight.ap import (APVersion, APSEL, APSEL_SHIFT, APv1Address)
from .stlink.usb import STLinkUSBInterface
from .stlink.stlink import STLink
from .stlink.detect.factory import create_mbed_detector
from ..board.mbed_board import MbedBoard
from ..board.board_ids import BOARD_ID_TO_INFO
from ..utility import conversion

if TYPE_CHECKING:
    from ..board.board_ids import BoardInfo

class StlinkProbe(DebugProbe):
    """@brief Wraps an STLink as a DebugProbe."""

    _board_id: Optional[str]

    # Shared cache for the STLink Mbed board IDs read from MSD volumes.
    # The dict maps the serial number to 4-character board ID, or None if no ID is available.
    _mbed_board_id_cache: Dict[str, Optional[str]] = {}

    @classmethod
    def get_all_connected_probes(cls, unique_id: Optional[str] = None,
            is_explicit: bool = False) -> List[StlinkProbe]:
        return [cls(dev) for dev in STLinkUSBInterface.get_all_connected_devices()]

    @classmethod
    def get_probe_with_id(cls, unique_id: str, is_explicit: bool = False) -> Optional[StlinkProbe]:
        for dev in STLinkUSBInterface.get_all_connected_devices():
            if dev.serial_number == unique_id:
                return cls(dev)
        return None

    def __init__(self, device: STLinkUSBInterface) -> None:
        super().__init__()
        self._link = STLink(device)
        self._is_open = False
        self._is_connected = False
        self._nreset_state = False
        self._memory_interfaces = {}
        self._board_id = None
        self._caps = set()

    @property
    def board_id(self) -> Optional[str]:
        """@brief Lazily loaded 4-character board ID."""
        if self._board_id is None:
            self._board_id = self._get_board_id()
        return self._board_id

    def _get_board_id(self) -> Optional[str]:
        # Try to get the board ID first by sending a command, since it is much faster. This requires
        # opening the USB device, however, and requires a recent STLink firmware version.
        board_id = self._link.get_board_id()
        if board_id is None:
            # Check the cache.
            if self._link.serial_number in self._mbed_board_id_cache:
                board_id = StlinkProbe._mbed_board_id_cache[self._link.serial_number]
            else:
                # Try to detect associated board info via the STLinkV2-1 MSD volume.
                detector = create_mbed_detector()
                if detector is not None:
                    for info in detector.list_mbeds():
                        usb_id = info['target_id_usb_id']

                        # Some STLink probes provide an MSD volume, but not the mbed.htm file.
                        # We can live without the board ID, so just ignore any error.
                        try:
                            this_board_id = info['target_id_mbed_htm'][0:4]
                        except KeyError:
                            # No board ID is available for this board.
                            StlinkProbe._mbed_board_id_cache[usb_id] = None
                        else:
                            # Populate the cache with the ID.
                            StlinkProbe._mbed_board_id_cache[usb_id] = this_board_id

                            # Use this ID if it's for our board.
                            if usb_id == self._link.serial_number:
                                board_id = this_board_id
                                break
        return board_id

    @property
    def description(self) -> str:
        return self.product_name

    @property
    def vendor_name(self):
        return self._link.vendor_name

    @property
    def product_name(self):
        return self._link.product_name

    @property
    def supported_wire_protocols(self):
        return [DebugProbe.Protocol.DEFAULT, DebugProbe.Protocol.SWD, DebugProbe.Protocol.JTAG]

    @property
    def unique_id(self):
        return self._link.serial_number

    @property
    def wire_protocol(self):
        return DebugProbe.Protocol.SWD if self._is_connected else None

    @property
    def is_open(self):
        return self._is_open

    @property
    def capabilities(self):
        return self._caps

    @property
    def associated_board_info(self) -> Optional[BoardInfo]:
        if (self.board_id is not None) and (self.board_id in BOARD_ID_TO_INFO):
            return BOARD_ID_TO_INFO[self.board_id]
        else:
            return None

    def create_associated_board(self):
        assert self.session is not None
        board_info = self.associated_board_info
        if board_info or self.board_id:
            return MbedBoard(self.session, board_info=board_info, board_id=self.board_id)
        else:
            return None

    def open(self):
        assert self.session is not None

        self._link.open()
        self._is_open = True

        # This call is ignored if the STLink is not V3.
        prescaler = self.session.options.get('stlink.v3_prescaler')
        if prescaler not in (1, 2, 4):
            prescaler = self.session.options.get_default('stlink.v3_prescaler')
        self._link.set_prescaler(prescaler)

        # Update capabilities.
        self._caps = {
                self.Capability.SWO,
                self.Capability.MANAGED_AP_SELECTION,
                self.Capability.MANAGED_DPBANKSEL,
                }
        if self._link.supports_banked_dp:
            self._caps.add(self.Capability.BANKED_DP_REGISTERS)

    def close(self):
        self._link.close()
        self._is_open = False

    # ------------------------------------------- #
    #          Target control functions
    # ------------------------------------------- #
    def connect(self, protocol=None):
        self._link.enter_debug(STLink.Protocol.SWD)
        self._is_connected = True

    def disconnect(self):
        # TODO Close the APs. When this is attempted, we get an undocumented 0x1d error. Doesn't
        #      seem to be necessary, anyway.
        self._memory_interfaces = {}

        self._link.enter_idle()
        self._is_connected = False

    def set_clock(self, frequency):
        self._link.set_swd_frequency(frequency)

    def reset(self):
        assert self.session
        self._link.drive_nreset(True)
        sleep(self.session.options.get('reset.hold_time'))
        self._link.drive_nreset(False)
        sleep(self.session.options.get('reset.post_delay'))

    def assert_reset(self, asserted):
        self._link.drive_nreset(asserted)
        self._nreset_state = asserted

    def is_reset_asserted(self):
        return self._nreset_state

    def flush(self):
        pass

    # ------------------------------------------- #
    #          DAP Access functions
    # ------------------------------------------- #

    def read_dp(self, addr, now=True):
        result = self._link.read_dap_register(STLink.DP_PORT, addr)

        def read_dp_result_callback():
            return result

        return result if now else read_dp_result_callback

    def write_dp(self, addr, data):
        self._link.write_dap_register(STLink.DP_PORT, addr, data)

    def read_ap(self, addr, now=True):
        apsel = (addr & APSEL) >> APSEL_SHIFT
        result = self._link.read_dap_register(apsel, addr & 0xffff)

        def read_ap_result_callback():
            return result

        return result if now else read_ap_result_callback

    def write_ap(self, addr, data):
        apsel = (addr & APSEL) >> APSEL_SHIFT
        self._link.write_dap_register(apsel, addr & 0xffff, data)

    def read_ap_multiple(self, addr, count=1, now=True):
        results = [self.read_ap(addr, now=True) for n in range(count)]

        def read_ap_multiple_result_callback():
            return results

        return results if now else read_ap_multiple_result_callback

    def write_ap_multiple(self, addr, values):
        for v in values:
            self.write_ap(addr, v)

    def get_memory_interface_for_ap(self, ap_address):
        assert self._is_connected
        # STLink memory access commands only support an 8-bit APSEL.
        if ap_address.ap_version != APVersion.APv1:
            return None
        assert isinstance(ap_address, APv1Address)
        apsel = ap_address.apsel
        if apsel not in self._memory_interfaces:
            self._link.open_ap(apsel)
            self._memory_interfaces[apsel] = STLinkMemoryInterface(self._link, apsel)
        return self._memory_interfaces[apsel]

    def swo_start(self, baudrate):
        self._link.swo_start(baudrate)

    def swo_stop(self):
        self._link.swo_stop()

    def swo_read(self):
        return self._link.swo_read()

class STLinkMemoryInterface(MemoryInterface):
    """@brief Concrete memory interface for a single AP."""

    def __init__(self, link, apsel):
        self._link = link
        self._apsel = apsel

    def write_memory(self, addr: int, data: int, transfer_size: int=32, **attrs: Any) -> None:
        """@brief Write a single memory location.

        By default the transfer size is a word.
        """
        assert transfer_size in (8, 16, 32)
        addr &= 0xffffffff
        csw = attrs.get('csw', 0)
        if transfer_size == 32:
            self._link.write_mem32(addr, conversion.u32le_list_to_byte_list([data]), self._apsel, csw)
        elif transfer_size == 16:
            self._link.write_mem16(addr, conversion.u16le_list_to_byte_list([data]), self._apsel, csw)
        elif transfer_size == 8:
            self._link.write_mem8(addr, [data], self._apsel, csw)

    def read_memory(self, addr: int, transfer_size: int=32, now: bool=True, **attrs: Any) \
            -> Union[int, Callable[[], int]]:
        """@brief Read a memory location.

        By default, a word will be read.
        """
        assert transfer_size in (8, 16, 32)
        addr &= 0xffffffff
        csw = attrs.get('csw', 0)
        if transfer_size == 32:
            result = conversion.byte_list_to_u32le_list(self._link.read_mem32(addr, 4, self._apsel, csw))[0]
        elif transfer_size == 16:
            result = conversion.byte_list_to_u16le_list(self._link.read_mem16(addr, 2, self._apsel, csw))[0]
        elif transfer_size == 8:
            result = self._link.read_mem8(addr, 1, self._apsel, csw)[0]

        def read_callback():
            return result
        return result if now else read_callback

    def write_memory_block32(self, addr: int, data: Sequence[int], **attrs: Any) -> None:
        addr &= 0xffffffff
        csw = attrs.get('csw', 0)
        self._link.write_mem32(addr, conversion.u32le_list_to_byte_list(data), self._apsel, csw)

    def read_memory_block32(self, addr: int, size: int, **attrs: Any) -> Sequence[int]:
        addr &= 0xffffffff
        csw = attrs.get('csw', 0)
        return conversion.byte_list_to_u32le_list(self._link.read_mem32(addr, size * 4, self._apsel, csw))

    def read_memory_block8(self, addr: int, size: int, **attrs: Any) -> Sequence[int]:
        addr &= 0xffffffff
        csw = attrs.get('csw', 0)
        res = []

        # Transfers are handled in 3 phases:
        #   1. read 8-bit chunks until the first aligned address is reached,
        #   2. read 32-bit chunks from all aligned addresses,
        #   3. read 8-bit chunks from the remaining unaligned addresses.
        # If the requested size is so small that phase-1 would not even reach
        # aligned address, go straight to phase-3.

        # 1. read leading unaligned bytes
        unaligned_count = 3 & (4 - addr)
        if (size > unaligned_count > 0):
            res += self._link.read_mem8(addr, unaligned_count, self._apsel, csw)
            size -= unaligned_count
            addr += unaligned_count

        # 2. read aligned block of 32 bits
        if (size >= 4):
            aligned_size = size & ~3
            res += self._link.read_mem32(addr, aligned_size, self._apsel, csw)
            size -= aligned_size
            addr += aligned_size

        # 3. read trailing unaligned bytes
        if (size > 0):
            res += self._link.read_mem8(addr, size, self._apsel, csw)

        return res

    def write_memory_block8(self, addr: int, data: Sequence[int], **attrs: Any) -> None:
        addr &= 0xffffffff
        csw = attrs.get('csw', 0)
        size = len(data)
        idx = 0

        # write leading unaligned bytes
        unaligned_count = 3 & (4 - addr)
        if (size > unaligned_count > 0):
            self._link.write_mem8(addr, data[:unaligned_count], self._apsel, csw)
            size -= unaligned_count
            addr += unaligned_count
            idx += unaligned_count

        # write aligned block of 32 bits
        if (size >= 4):
            aligned_size = size & ~3
            self._link.write_mem32(addr, data[idx:idx + aligned_size], self._apsel, csw)
            size -= aligned_size
            addr += aligned_size
            idx += aligned_size

        # write trailing unaligned bytes
        if (size > 0):
            self._link.write_mem8(addr, data[idx:], self._apsel, csw)

class StlinkProbePlugin(Plugin):
    """@brief Plugin class for StlLinkProbe."""

    def should_load(self):
        # TODO only load the plugin when libusb is available
        return True

    def load(self):
        return StlinkProbe

    @property
    def name(self):
        return "stlink"

    @property
    def description(self):
        return "STMicro STLinkV2 and STLinkV3 debug probe"

    @property
    def options(self) -> List[OptionInfo]:
        return [
            OptionInfo('stlink.v3_prescaler', int, 1,
                    "Sets the HCLK prescaler of an STLinkV3, changing performance versus power tradeoff. "
                    "The value must be one of 1=high performance (default), 2=normal, or 4=low power.")
        ]



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import io
from typing import List

from .base import SubcommandBase
from ..commands.commander import PyOCDCommander

class CommanderSubcommand(SubcommandBase):
    """@brief `pyocd commander` subcommand."""

    NAMES = ['commander', 'cmd']
    HELP = "Interactive command console."
    DEFAULT_LOG_LEVEL = logging.WARNING

    EPILOG = """Commands specified by the -c/--command and -x/--execute arguments are run in the order they are listed
        on the command line, and the two types can be mixed freely and in any order. Normally, pyOCD will exit after
        running such commands. If the -i/--interactive flag is set, then the interactive REPL will be instead be
        started when the commands have finished. In command files each line is either a complete command, a comment
        started with '#', or empty.
        """

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        commander_parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        commander_options = commander_parser.add_argument_group("commander options")
        commander_options.add_argument("-H", "--halt", action="store_true", default=None,
            help="Halt core upon connect. (Deprecated, see --connect.)")
        commander_options.add_argument("-N", "--no-init", action="store_true",
            help="Do not init debug system.")
        commander_options.add_argument("--elf", metavar="PATH",
            help="Optionally specify ELF file being debugged.")
        commander_options.add_argument("-c", "--command", dest="commands", metavar="CMD", action='append', nargs='+',
            help="Run commands.")
        commander_options.add_argument("-x", "--execute", dest="commands", metavar="FILE", action='append',
            type=argparse.FileType('r'),
            help="Execute commands from file. Pass - for stdin.")
        commander_options.add_argument("-i", "--interactive", action="store_true",
            help="Stay in interactive mode after running commands specified from command line or file.")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, commander_parser]

    def invoke(self) -> int:
        """@brief Handle 'commander' subcommand."""
        # Flatten commands list then extract primary command and its arguments.
        if self._args.commands is not None:
            cmds = []
            for cmd in self._args.commands:
                if isinstance(cmd, io.IOBase):
                    cmds.append(cmd)
                else:
                    cmds.append(" ".join(cmd))
        else:
            cmds = None

        # Enter REPL.
        PyOCDCommander(self._args, cmds).run()

        return 0





# pyOCD debugger
# Copyright (c) 2018 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile

class DBGMCU:
    # en.DM00108282.pdf (STM32L0x1 Reference manual)
    # 27.9.3 Debug MCU configuration register (DBG_CR)
    DBG_CR = 0x40015804
    DBG_STANDBY = (1 << 2)
    DBG_STOP = (1 << 1)
    DBG_SLEEP = 1
    DBG_CR_VALUE = DBG_STANDBY | DBG_STOP | DBG_SLEEP

    DBG_APB1_FZ = 0x40015808
    DBG_APB1_FZ_VALUE = 0b11000000001000000001110000110011

    DBG_APB2_FZ = 0x4001580C
    DBG_APB2_FZ_VALUE = 0b100100


FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0xd0012a01, 0xd1172a02, 0x6981483b, 0x0212220f, 0x61814311, 0x60c14939, 0x60c14939, 0x61014939,
    0x61014939, 0x02c069c0, 0x4839d406, 0x60014937, 0x60412106, 0x60814937, 0x47702000, 0xd0012801,
    0xd1082802, 0x6841482c, 0x43112202, 0x68416041, 0x43112201, 0x20006041, 0xb5304770, 0x684a4926,
    0x4322154c, 0x684a604a, 0x432a2508, 0x2200604a, 0x48296002, 0xe0004a26, 0x698b6010, 0xd1fb07db,
    0x43a06848, 0x68486048, 0x604843a8, 0xbd302000, 0x47702001, 0x4c18b5f0, 0x15252300, 0x313f2608,
    0x468c0989, 0x6861e024, 0x60614329, 0x43316861, 0x21406061, 0xc080ca80, 0x29001f09, 0x4916d1fa,
    0x07ff69a7, 0x4f12d002, 0xe7f96039, 0x050969a1, 0xd0060f09, 0x210f69a0, 0x43080209, 0x200161a0,
    0x6861bdf0, 0x606143a9, 0x43b16861, 0x1c5b6061, 0xd8d8459c, 0xbdf02000, 0x40022000, 0x89abcdef,
    0x02030405, 0x8c9daebf, 0x13141516, 0x00005555, 0x40003000, 0x00000fff, 0x0000aaaa, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000021,
    'pc_unInit': 0x2000005d,
    'pc_program_page': 0x200000b5,
    'pc_erase_sector': 0x2000007b,
    'pc_eraseAll': 0x12000001,

    'static_base' : 0x20000000 + 0x00000020 + 0x0000011c,
    'begin_stack' : 0x20000400,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x400,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000],
    'min_program_length' : 0x400,

    # Flash information
    'flash_start': 0x8000000,
    'flash_size': 0x8000,
    'sector_sizes': (
        (0x0, 0x80),
    )
}


class STM32L031x6(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        FlashRegion(name='Flash', start=0x08000000, length=0x8000, blocksize=0x80,  is_boot_memory=True, algo=FLASH_ALGO),
        RamRegion(name='RAM', start=0x20000000, length=0x2000),
        FlashRegion(name='EEPROM', start=0x08080000, length=0x400, blocksize=0x400, algo=FLASH_ALGO)
        )

    def __init__(self, session):
        super(STM32L031x6, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("STM32L0x1.svd")

    def post_connect_hook(self):
        self.write32(DBGMCU.DBG_CR, DBGMCU.DBG_CR_VALUE)
        self.write32(DBGMCU.DBG_APB1_FZ, DBGMCU.DBG_APB1_FZ_VALUE)
        self.write32(DBGMCU.DBG_APB2_FZ, DBGMCU.DBG_APB2_FZ_VALUE)



# pyOCD debugger
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import prettytable
from typing import (Any, Dict, List, Optional, Type)

from ..utility.cmdline import convert_frequency

class SubcommandBase:
    """@brief Base class for pyocd command line subcommand."""

    # Subcommand descriptors.
    NAMES: List[str] = []
    HELP: str = ""
    EPILOG: Optional[str] = None
    DEFAULT_LOG_LEVEL = logging.INFO
    SUBCOMMANDS: List[Type["SubcommandBase"]] = []

    ## Class attribute to store the built subcommand argument parser.
    parser: Optional[argparse.ArgumentParser] = None

    class CommonOptions:
        """@brief Namespace with parsers for repeated option groups."""

        # Define logging related options.
        LOGGING = argparse.ArgumentParser(description='logging', add_help=False)
        LOGGING_GROUP = LOGGING.add_argument_group("logging")
        LOGGING_GROUP.add_argument('-v', '--verbose', action='count', default=0,
            help="Increase logging level. Can be specified multiple times.")
        LOGGING_GROUP.add_argument('-q', '--quiet', action='count', default=0,
            help="Decrease logging level. Can be specified multiple times.")
        LOGGING_GROUP.add_argument('-L', '--log-level', action='append', metavar="LOGGERS=LEVEL", default=[],
            help="Set log level of loggers whose name matches any of the comma-separated list of glob-style "
            "patterns. Log level must be one of (critical, error, warning, info, debug). Can be "
            "specified multiple times. Example: -L*.trace,pyocd.core.*=debug")
        LOGGING_GROUP.add_argument('--color', choices=("always", "auto", "never"), default=None, nargs='?',
            const="auto", help="Control color logging. Default is auto.")

        # Define config related options for all subcommands.
        CONFIG = argparse.ArgumentParser(description='common', add_help=False)
        CONFIG_GROUP = CONFIG.add_argument_group("configuration")
        CONFIG_GROUP.add_argument('-j', '--project', '--dir', metavar="PATH", dest="project_dir",
            help="Set the project directory. Defaults to the directory where pyocd was run.")
        CONFIG_GROUP.add_argument('--config', metavar="PATH",
            help="Specify YAML configuration file. Defaults to pyocd.yaml or pyocd.yml in the project directory.")
        CONFIG_GROUP.add_argument("--no-config", action="store_true", default=None,
            help="Do not use a configuration file.")
        CONFIG_GROUP.add_argument('--script', metavar="PATH",
            help="Use the specified user script. Defaults to pyocd_user.py in the project directory.")
        CONFIG_GROUP.add_argument('-O', action='append', dest='options', metavar="OPTION=VALUE",
            help="Set named option.")
        CONFIG_GROUP.add_argument("-da", "--daparg", dest="daparg", nargs='+',
            help="(Deprecated) Send setting to DAPAccess layer.")
        CONFIG_GROUP.add_argument("--pack", metavar="PATH", action="append",
            help="Path to the .pack file for a CMSIS Device Family Pack.")

        # Define common options for all subcommands, including logging options.
        COMMON = argparse.ArgumentParser(description='common',
            parents=[LOGGING, CONFIG], add_help=False)

        # Common connection related options.
        CONNECT = argparse.ArgumentParser(description='common', add_help=False)
        CONNECT_GROUP = CONNECT.add_argument_group("connection")
        CONNECT_GROUP.add_argument("-u", "--uid", "--probe", dest="unique_id",
            help="Select the debug probe by its full or partial unique ID. Optionally prefixed with "
            "'<probe-type>:' where <probe-type> is the name of a probe plugin.")
        # Reserved --board argument.
        # CONNECT_GROUP.add_argument("-b", "--board", dest="board_override", metavar="BOARD",
        #     help="Set the board type (not yet implemented).")
        CONNECT_GROUP.add_argument("-t", "--target", dest="target_override", metavar="TARGET",
            help="Set the target type. See available target types with 'pyocd list --targets'.")
        CONNECT_GROUP.add_argument("-f", "--frequency", dest="frequency", default=None, type=convert_frequency,
            help="SWD/JTAG clock frequency in Hz. Accepts a float or int with optional case-"
                "insensitive K/M suffix and optional Hz. Examples: \"1000\", \"2.5khz\", \"10m\".")
        CONNECT_GROUP.add_argument("-W", "--no-wait", action="store_true",
            help="Do not wait for a probe to be connected if none are available.")
        CONNECT_GROUP.add_argument("-M", "--connect", dest="connect_mode", metavar="MODE",
            help="Select connect mode from one of (halt, pre-reset, under-reset, attach).")

    @classmethod
    def add_subcommands(cls, parser: argparse.ArgumentParser) -> None:
        """@brief Add declared subcommands to the given parser."""
        if cls.SUBCOMMANDS:
            subparsers = parser.add_subparsers(title="subcommands", metavar="", dest='cmd')
            for subcmd_class in cls.SUBCOMMANDS:
                parsers = subcmd_class.get_args()
                subcmd_class.parser = parsers[-1]

                subparser = subparsers.add_parser(
                                subcmd_class.NAMES[0],
                                aliases=subcmd_class.NAMES[1:],
                                parents=parsers,
                                help=subcmd_class.HELP,
                                epilog=subcmd_class.EPILOG)
                subparser.set_defaults(command_class=subcmd_class)
                subcmd_class.customize_subparser(subparser)

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object.
        @return List of argument parsers. The last element in the list _must_ be the parser for the subcommand
            class itself, as it is saved by the caller in cls.parser.
        """
        raise NotImplementedError()

    @classmethod
    def customize_subparser(cls, subparser: argparse.ArgumentParser) -> None:
        """@brief Optionally modify a subparser after it is created."""
        pass

    def __init__(self, args: argparse.Namespace):
        """@brief Constructor.

        @param self This object.
        @param args Namespace of parsed argument values.
        """
        self._args = args

    def invoke(self) -> int:
        """@brief Run the subcommand.
        @return Process status code for the command.
        """
        if self.parser is not None:
            self.parser.print_help()
        return 0

    def _get_log_level_delta(self) -> int:
        """@brief Compute the logging level delta sum from quiet and verbose counts."""
        return (self._args.quiet * 10) - (self._args.verbose * 10)

    def _increase_logging(self, loggers: List[str]) -> None:
        """@brief Increase logging level for a set of subloggers.
        @param self This object.
        @param loggers
        """
        delta = self._get_log_level_delta()
        if delta <= 0:
            level = max(1, self.DEFAULT_LOG_LEVEL + delta - 10)
            for logger in loggers:
                logging.getLogger(logger).setLevel(level)

    def _get_pretty_table(self, fields: List[str], header: bool = None) -> prettytable.PrettyTable:
        """@brief Returns a PrettyTable object with formatting options set."""
        pt = prettytable.PrettyTable(fields)
        pt.align = 'l'
        if header is not None:
            pt.header = header
        elif hasattr(self._args, 'no_header'):
            pt.header = not self._args.no_header
        else:
            pt.header = True
        pt.border = True
        pt.hrules = prettytable.HEADER
        pt.vrules = prettytable.NONE
        return pt

    def _modified_option_defaults(self) -> Dict[str, Any]:
        """@brief Returns a dict of session option defaults.

        @return A dict containing updated default values for session options, based on common
            subcommand arguments. It is intended to be passed as the `option_defaults` argument when
            creating a `Session` instance.
        @precondition Logging must have been configured.
        """
        return {
            # Change 'debug.traceback' default to True if debug logging is enabled.
            'debug.traceback': logging.getLogger('pyocd').isEnabledFor(logging.DEBUG),
        }





# pyOCD debugger
# Copyright (c) 2023 David van Rijn
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class MinimalMemAP:
    """
    Minimalistic Access Port implementation.
    This is used in some targets to access memory before the "real" AP's become available.
    """
    AP0_CSW_ADDR = 0x00
    AP0_CSW_ADDR_VAL = 0x03000012
    AP0_TAR_ADDR = 0x04
    AP0_IDR_ADDR = 0xFC
    AP0_DRW_ADDR = 0x0C

    def __init__(self, dp):
        self.dp = dp

    def init(self):
        # Init AP #0
        IDR = self.dp.read_ap(MinimalMemAP.AP0_IDR_ADDR)
        # Check expected MEM-AP
        assert IDR&0x0fffe00f == 0x04770001, f"Wrong IDR read from device: 0x{IDR:08x}"
        self.dp.write_ap(MinimalMemAP.AP0_CSW_ADDR, MinimalMemAP.AP0_CSW_ADDR_VAL)

    def read32(self, addr):
        self.dp.write_ap(MinimalMemAP.AP0_TAR_ADDR, addr)
        return self.dp.read_ap(MinimalMemAP.AP0_DRW_ADDR)

    def write32(self, addr, val):
        self.dp.write_ap(MinimalMemAP.AP0_TAR_ADDR, addr)
        self.dp.write_ap(MinimalMemAP.AP0_DRW_ADDR, val)



# pyOCD debugger
# Copyright (c) 2017-2019 Arm Limited
# COpyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections.abc
from typing import (TYPE_CHECKING, Iterable, List, Optional, Sequence, Union)

if TYPE_CHECKING:
    from .events import TraceEvent

class TraceEventSink:
    """@brief Abstract interface for a trace event sink."""
    def receive(self, event: "TraceEvent") -> None:
        """@brief Handle a single trace event.
        @param self
        @param event An instance of TraceEvent or one of its subclasses.
        """
        raise NotImplementedError()

class TraceEventFilter(TraceEventSink):
    """@brief Abstract interface for a trace event filter."""

    def __init__(self, sink: Optional[TraceEventSink] = None) -> None:
        self._sink = sink

    def connect(self, sink: TraceEventSink) -> None:
        """@brief Connect the downstream trace sink or filter."""
        self._sink = sink

    def receive(self, event: "TraceEvent") -> None:
        """@brief Handle a single trace event.

        Passes the event through the filter() method. If one or more objects are returned, they
        are then passed to the trace sink connected to this filter (which may be another filter).

        @param self
        @param event An instance of TraceEvent or one of its subclasses.
        """
        filtered_event = self.filter(event)
        if (filtered_event is not None) and (self._sink is not None):
            if isinstance(event, collections.abc.Iterable):
                for event_item in event:
                    self._sink.receive(event_item)
            else:
                self._sink.receive(event)

    def filter(self, event: "TraceEvent") -> Union[None, "TraceEvent", Sequence["TraceEvent"]]:
        """@brief Filter a single trace event.

        @param self
        @param event An instance of TraceEvent or one of its subclasses.
        @return Either None, a single TraceEvent, or a sequence of TraceEvents.
        """
        raise NotImplementedError()

class TraceEventTee(TraceEventSink):
    """@brief Trace event sink that replicates events to multiple sinks."""

    def __init__(self) -> None:
        self._sinks: List[TraceEventSink] = []

    def connect(self, sinks: Iterable[TraceEventSink]) -> None:
        """@brief Connect one or more downstream trace sinks.

        @param self
        @param sinks If this parameter is a single object, it will be added to the list of
          downstream trace event sinks. If it is an iterable (list, tuple, etc.), then it will
          completely replace the current list of trace event sinks.
        """
        if isinstance(sinks, collections.abc.Iterable):
            self._sinks = list(sinks)
        elif sinks not in self._sinks:
            self._sinks.append(sinks)

    def receive(self, event: "TraceEvent") -> None:
        """@brief Replicate a single trace event to all connected downstream trace event sinks.

        @param self
        @param event An instance of TraceEvent or one of its subclasses.
        """
        for sink in self._sinks:
            sink.receive(event)




# Copyright (c) 2018-2019, Arm Limited and affiliates.
# Copyright (c) 2021 Chris Reed.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
from collections import defaultdict
from copy import copy
import logging

from .base import StlinkDetectBase

LOG = logging.getLogger(__name__)

try:
    if sys.version_info[0] < 3:
        import _winreg as winreg
    else:
        import winreg
except ImportError:
    # Only allow the exception to propagate on Windows. This module should never be imported
    # in the first place on other OSes, allowing imports is simply for easier testing.
    import platform
    if platform.system() == "Windows":
        raise

MAX_COMPOSITE_DEVICE_SUBDEVICES = 5
MBED_STORAGE_DEVICE_VENDOR_STRINGS = [
    "ven_mbed",
]


def _get_values_with_numeric_keys(reg_key):
    result = []
    try:
        for v in _iter_vals(reg_key):
            try:
                # The only values we care about are ones that have an integer key.
                # The other values are metadata for the registry
                int(v[0])
                result.append(v[1])
            except ValueError:
                continue
    except OSError:
        LOG.debug("Failed to iterate over all keys")

    return result


def _is_mbed_volume(volume_string):
    for vendor_string in MBED_STORAGE_DEVICE_VENDOR_STRINGS:
        if vendor_string.lower() in volume_string.lower():
            return True

    return False


def _get_cached_mounted_points():
    """Get the volumes present on the system
    @return List of mount points and their associated target id
      Ex. [{ 'mount_point': 'D:', 'target_id_usb_id': 'xxxx'}, ...]
    """
    result = []
    try:
        # Open the registry key for mounted devices
        mounted_devices_key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, "SYSTEM\\MountedDevices"
        )
        for v in _iter_vals(mounted_devices_key):
            # Valid entries have the following format: \DosDevices\D:
            if "DosDevices" not in v[0]:
                continue

            volume_string = v[1].decode("utf-16le", "ignore")
            if not _is_mbed_volume(volume_string):
                continue

            mount_point_match = re.match(".*\\\\(.:)$", v[0])

            if not mount_point_match:
                LOG.debug("Invalid disk pattern for entry %s, skipping", v[0])
                continue

            mount_point = mount_point_match.group(1)

            result.append({"mount_point": mount_point, "volume_string": volume_string})
    except OSError:
        LOG.error('Failed to open "MountedDevices" in registry')

    return result


def _get_disks():
    LOG.debug("Fetching mounted devices from disk service registry entry")
    try:
        disks_key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Services\\Disk\\Enum"
        )
        disk_strings = _get_values_with_numeric_keys(disks_key)
        return [v for v in disk_strings if _is_mbed_volume(v)]
    except OSError:
        LOG.debug("No disk service found, no device can be detected")
        return []


def _get_usb_storage_devices():
    LOG.debug("Fetching usb storage devices from USBSTOR service registry entry")
    try:
        usbstor_key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            "SYSTEM\\CurrentControlSet\\Services\\USBSTOR\\Enum",
        )
        return _get_values_with_numeric_keys(usbstor_key)
    except OSError:
        LOG.debug("No USBSTOR service found, no device can be detected")
        return []


def _determine_valid_non_composite_devices(devices, target_id_usb_id_mount_point_map):
    # Some Mbed devices do not expose a composite USB device. This is typical for
    # DAPLink devices in bootloader mode. Since we only have to check one endpoint
    # (specifically, the mass storage device), we handle this case separately
    candidates = {}
    for device in devices:
        device_key_string = "SYSTEM\\CurrentControlSet\\Enum\\" + device["full_path"]
        try:
            device_key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, device_key_string)
        except OSError:
            LOG.debug('Key "%s" not found', device_key_string)
            continue

        try:
            capability = _determine_subdevice_capability(device_key)
        except CompatibleIDsNotFoundException:
            LOG.debug(
                'Expected %s to have subkey "CompatibleIDs". Skipping.',
                device_key_string,
            )
            continue

        if capability != "msd":
            LOG.debug(
                "Expected msd device but got %s, skipping %s",
                capability,
                device["full_path"],
            )
            continue

        target_id_usb_id = device["entry_key_string"]
        try:
            candidates[target_id_usb_id] = {
                "target_id_usb_id": target_id_usb_id,
                "mount_point": target_id_usb_id_mount_point_map[target_id_usb_id],
            }

            candidates[target_id_usb_id].update(
                _vid_pid_path_to_usb_info(device["vid_pid_path"])
            )
        except KeyError:
            pass

    return candidates


def _determine_subdevice_capability(key):
    try:
        vals = winreg.QueryValueEx(key, "CompatibleIDs")
        compatible_ids = [x.lower() for x in vals[0]]
    except OSError:
        raise CompatibleIDsNotFoundException()

    if "usb\\class_00" in compatible_ids or "usb\\devclass_00" in compatible_ids:
        return "composite"
    elif "usb\\class_08" in compatible_ids:
        return "msd"
    elif "usb\\class_02" in compatible_ids:
        return "serial"
    else:
        LOG.debug("Unknown capabilities from the following ids: %s", compatible_ids)
        return None


def _vid_pid_path_to_usb_info(vid_pid_path):
    """Provide the vendor ID and product ID of a device based on its entry in the registry
    @return Returns {'vendor_id': '<vendor ID>', 'product': '<product ID>'}
    @details If the vendor ID or product ID can't be determined, they will be returned
    as None.
    """
    result = {"vendor_id": None, "product_id": None}

    for component in vid_pid_path.split("&"):
        component_part = component.lower().split("_")

        if len(component_part) != 2:
            LOG.debug("Unexpected VID/PID string structure %s", component)
            break

        if component_part[0] == "vid":
            result["vendor_id"] = component_part[1]
        elif component_part[0] == "pid":
            result["product_id"] = component_part[1]

    return result


def _iter_keys_as_str(key):
    """Iterate over subkeys of a key returning subkey as string
    """
    for i in range(winreg.QueryInfoKey(key)[0]):
        yield winreg.EnumKey(key, i)


def _iter_keys(key):
    """Iterate over subkeys of a key
    """
    for i in range(winreg.QueryInfoKey(key)[0]):
        yield winreg.OpenKey(key, winreg.EnumKey(key, i))


def _iter_vals(key):
    """Iterate over values of a key
    """
    for i in range(winreg.QueryInfoKey(key)[1]):
        yield winreg.EnumValue(key, i)


class CompatibleIDsNotFoundException(Exception):
    pass


class StlinkDetectWindows(StlinkDetectBase):
    """ mbed-enabled platform detection for Windows
    """

    def __init__(self, **kwargs):
        StlinkDetectBase.__init__(self, **kwargs)

    def find_candidates(self):
        cached_mount_points = _get_cached_mounted_points()
        disks = _get_disks()
        usb_storage_devices = _get_usb_storage_devices()

        target_id_usb_id_mount_point_map = {}
        for cached_mount_point_info in cached_mount_points:
            for index, disk in enumerate(copy(disks)):
                match_string = disk.split("\\")[-1]
                if match_string in cached_mount_point_info["volume_string"]:
                    # TargetID is a hex string with 10-48 chars
                    target_id_usb_id_match = re.search(
                        "[&#]([0-9A-Za-z]{10,48})[&#]",
                        cached_mount_point_info["volume_string"],
                    )
                    if not target_id_usb_id_match:
                        LOG.debug(
                            "Entry %s has invalid target id pattern %s, skipping",
                            cached_mount_point_info["mount_point"],
                            cached_mount_point_info["volume_string"],
                        )
                        continue

                    target_id_usb_id_mount_point_map[
                        target_id_usb_id_match.group(1)
                    ] = cached_mount_point_info["mount_point"]
                    disks.pop(index)
                    break

        LOG.debug(
            "target_id_usb_id -> mount_point mapping: %s ",
            target_id_usb_id_mount_point_map,
        )
        non_composite_devices = []
        composite_devices = []
        for vid_pid_path in usb_storage_devices:
            # Split paths like "USB\VID_0483&PID_374B&MI_01\7&25b4dc8e&0&0001" by "\"
            vid_pid_path_componets = vid_pid_path.split("\\")

            vid_pid_components = vid_pid_path_componets[1].split("&")

            if len(vid_pid_components) != 2 and len(vid_pid_components) != 3:
                LOG.debug(
                    "Skipping USBSTOR device with unusual VID/PID string format '%s'",
                    vid_pid_path,
                )
                continue

            device = {
                "full_path": vid_pid_path,
                "vid_pid_path": "&".join(vid_pid_components[:2]),
                "entry_key_string": vid_pid_path_componets[2],
            }

            # A composite device's vid/pid path always has a third component
            if len(vid_pid_components) == 3:
                composite_devices.append(device)
            else:
                non_composite_devices.append(device)

        candidates = defaultdict(dict)
        candidates.update(
            _determine_valid_non_composite_devices(
                non_composite_devices, target_id_usb_id_mount_point_map
            )
        )
        # Now we'll find all valid VID/PID and target ID combinations
        target_id_usb_ids = set(target_id_usb_id_mount_point_map.keys()) - set(
            candidates.keys()
        )
        vid_pid_entry_key_string_map = defaultdict(set)

        for device in composite_devices:
            vid_pid_entry_key_string_map[device["vid_pid_path"]].add(
                device["entry_key_string"]
            )

        vid_pid_target_id_usb_id_map = defaultdict(dict)
        usb_key_string = "SYSTEM\\CurrentControlSet\\Enum\\USB"
        for vid_pid_path, entry_key_strings in vid_pid_entry_key_string_map.items():
            vid_pid_key_string = "%s\\%s" % (usb_key_string, vid_pid_path)
            try:
                vid_pid_key = winreg.OpenKey(
                    winreg.HKEY_LOCAL_MACHINE, vid_pid_key_string
                )
                target_id_usb_id_sub_keys = set(
                    [k for k in _iter_keys_as_str(vid_pid_key)]
                )
            except OSError:
                LOG.debug('VID/PID "%s" not found', vid_pid_key_string)
                continue

            overlapping_target_id_usb_ids = target_id_usb_id_sub_keys.intersection(
                set(target_id_usb_ids)
            )
            for target_id_usb_id in overlapping_target_id_usb_ids:
                composite_device_key_string = "%s\\%s" % (
                    vid_pid_key_string,
                    target_id_usb_id,
                )
                composite_device_key = winreg.OpenKey(vid_pid_key, target_id_usb_id)

                entry_key_string = target_id_usb_id
                is_prefix = False

                try:
                    new_entry_key_string, _ = winreg.QueryValueEx(
                        composite_device_key, "ParentIdPrefix"
                    )

                    if any(
                        e.startswith(new_entry_key_string) for e in entry_key_strings
                    ):
                        entry_key_string = new_entry_key_string
                        is_prefix = True
                except OSError:
                    LOG.debug(
                        'Device %s did not have a "ParentIdPrefix" key, '
                        "sticking with %s as entry key string",
                        composite_device_key_string,
                        target_id_usb_id,
                    )

                vid_pid_target_id_usb_id_map[vid_pid_path][entry_key_string] = {
                    "target_id_usb_id": target_id_usb_id,
                    "is_prefix": is_prefix,
                }

        for (
            vid_pid_path,
            entry_key_string_target_id_usb_id_map,
        ) in vid_pid_target_id_usb_id_map.items():
            for composite_device_subdevice_number in range(
                MAX_COMPOSITE_DEVICE_SUBDEVICES
            ):
                subdevice_type_key_string = "%s\\%s&MI_0%d" % (
                    usb_key_string,
                    vid_pid_path,
                    composite_device_subdevice_number,
                )
                try:
                    subdevice_type_key = winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE, subdevice_type_key_string
                    )
                except OSError:
                    LOG.debug(
                        "Composite device subdevice key %s was not found, skipping",
                        subdevice_type_key_string,
                    )
                    continue

                for (
                    entry_key_string,
                    entry_data,
                ) in entry_key_string_target_id_usb_id_map.items():
                    if entry_data["is_prefix"]:
                        prepared_entry_key_string = "%s&000%d" % (
                            entry_key_string,
                            composite_device_subdevice_number,
                        )
                    else:
                        prepared_entry_key_string = entry_key_string
                    subdevice_key_string = "%s\\%s" % (
                        subdevice_type_key_string,
                        prepared_entry_key_string,
                    )
                    try:
                        subdevice_key = winreg.OpenKey(
                            subdevice_type_key, prepared_entry_key_string
                        )
                    except OSError:
                        LOG.debug(
                            "Sub-device %s not found, skipping", subdevice_key_string
                        )
                        continue

                    try:
                        capability = _determine_subdevice_capability(subdevice_key)
                    except CompatibleIDsNotFoundException:
                        LOG.debug(
                            'Expected %s to have subkey "CompatibleIDs". Skipping.',
                            subdevice_key_string,
                        )
                        continue

                    if capability == "msd":
                        candidates[entry_data["target_id_usb_id"]][
                            "mount_point"
                        ] = target_id_usb_id_mount_point_map[
                            entry_data["target_id_usb_id"]
                        ]
                        candidates[entry_data["target_id_usb_id"]].update(
                            _vid_pid_path_to_usb_info(vid_pid_path)
                        )
                    elif capability == "serial":
                        try:
                            device_parameters_key = winreg.OpenKey(
                                subdevice_key, "Device Parameters"
                            )
                        except OSError:
                            LOG.debug(
                                'Key "Device Parameters" not under serial device entry'
                            )
                            continue

                        try:
                            candidates[entry_data["target_id_usb_id"]][
                                "serial_port"
                            ], _ = winreg.QueryValueEx(
                                device_parameters_key, "PortName"
                            )
                            candidates[entry_data["target_id_usb_id"]].update(
                                _vid_pid_path_to_usb_info(vid_pid_path)
                            )
                        except OSError:
                            LOG.debug(
                                '"PortName" value not found under serial device entry'
                            )
                            continue

        final_candidates = []
        for target_id_usb_id, candidate in candidates.items():
            candidate["target_id_usb_id"] = target_id_usb_id

            if "serial_port" not in candidate:
                candidate["serial_port"] = None

            if "mount_point" not in candidate:
                candidate["mount_point"] = None

            final_candidates.append(candidate)

        return final_candidates

    def mount_point_ready(self, path):
        """Check if a mount point is ready for file operations
        @return Returns True if the given path exists, False otherwise
        @details Calling the Windows command `dir` instead of using the python
        `os.path.exists`. The latter causes a Python error box to appear claiming
        there is "No Disk" for some devices that are in the ejected state. Calling
        `dir` prevents this since it uses the Windows API to determine if the
        device is ready before accessing the file system.
        """
        stdout, stderr, retcode = self._run_cli_process("dir %s" % path)
        result = True if retcode == 0 else False

        return result



# pyOCD debugger
# Copyright (c) 2020-2021 Federico Zuccardi Merli
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .provider import (TargetThread, ThreadProvider)
from .common import (read_c_string, HandlerModeThread,
                     EXC_RETURN_EXT_FRAME_MASK)
from ..core import exceptions
from ..core.target import Target
from ..core.plugin import Plugin
from ..debug.context import DebugContext
from ..coresight.cortex_m_core_registers import index_for_reg
from ..coresight.core_ids import CoreArchitecture

import logging

TX_THREAD_ID = 0x54485244  # 'THRD'
THREAD_ID_OFFSET = 0
THREAD_STACK_POINTER_OFFSET = 8
# All the following offset may be messed up if thread extensions are defined.
# They should be read somehow from the elf.
THREAD_NAME_OFFSET = 40
THREAD_PRIORITY_OFFSET = 44
THREAD_STATE_OFFSET = 48
THREAD_NEXT_OFFSET = 136

# Create a logger for this module.
LOG = logging.getLogger(__name__)


class TargetList(object):
    def __init__(self, context, ptr):
        self._context = context
        self._list = ptr

    def __iter__(self):
        next = 0
        head = self._context.read32(self._list)
        node = head
        is_valid = head != 0

        while is_valid and next != head:
            try:
                # Check if this is really a thread
                if self._context.read32(node) == TX_THREAD_ID:
                    # Yields the thread pointer.
                    yield node
                else:
                    # Something is wrong. Might depend on a thread extension
                    is_valid = False
                    LOG.warning(
                        "Wrong thread ID found. Memory corruption or unknown extensions")

                next = self._context.read32(node + THREAD_NEXT_OFFSET)
                node = next
            except exceptions.TransferError:
                LOG.warning(
                    "TransferError while reading list elements (list=0x%08x, node=0x%08x), terminating list", self._list, node)
                is_valid = False


class ThreadXThreadContext(DebugContext):
    """@brief Thread context for ThreadX."""

    # SP/PSP are handled specially, so it is not in these dicts.

    # Offsets are relative to stored SP in a task switch block, for the
    # combined software + hardware stacked registers. In exception case,
    # software registers are not stacked, so appropriate amount must be
    # subtracted.
    NOFPU_REGISTER_OFFSETS = {
        # Software stacked
        -1:     0,      # lr (exception)
        4:      4,      # r4
        5:      8,      # r5
        6:      12,     # r6
        7:      16,     # r7
        8:      20,     # r8
        9:      24,     # r9
        10:     28,     # r10
        11:     32,     # r11
        # Hardware stacked
        0:      36,     # r0
        1:      40,     # r1
        2:      44,     # r2
        3:      48,     # r3
        12:     52,     # r12
        14:     56,     # lr (thread)
        15:     60,     # pc
        16:     64,     # xpsr
    }

    # Cortex-m0 port of threadx reverses r4-7 and r8-11
    NOFPU_REGISTER_OFFSETS_V6M = {
        # Software stacked
        4:      20,     # r4
        5:      24,     # r5
        6:      28,     # r6
        7:      32,     # r7
        8:      4,      # r8
        9:      8,      # r9
        10:     12,     # r10
        11:     16,     # r11
    }

    FPU_REGISTER_OFFSETS = {
        # Software stacked
        -1:     0,      # lr (exception)
        0x50:   4,      # s16
        0x51:   8,      # s17
        0x52:   12,     # s18
        0x53:   16,     # s19
        0x54:   20,     # s20
        0x55:   24,     # s21
        0x56:   28,     # s22
        0x57:   32,     # s23
        0x58:   36,     # s24
        0x59:   40,     # s25
        0x5a:   44,     # s26
        0x5b:   48,     # s27
        0x5c:   52,     # s28
        0x5d:   56,     # s29
        0x5e:   60,     # s30
        0x5f:   64,     # s31
        4:      68,     # r4
        5:      72,     # r5
        6:      76,     # r6
        7:      80,     # r7
        8:      84,     # r8
        9:      88,     # r9
        10:     92,     # r10
        11:     96,     # r11
        # Hardware stacked
        0:      100,    # r0
        1:      104,    # r1
        2:      108,    # r2
        3:      112,    # r3
        12:     116,    # r12
        14:     120,    # lr
        15:     124,    # pc
        16:     128,    # xpsr
        0x40:   132,    # s0
        0x41:   136,    # s1
        0x42:   140,    # s2
        0x43:   144,    # s3
        0x44:   148,    # s4
        0x45:   152,    # s5
        0x46:   156,    # s6
        0x47:   160,    # s7
        0x48:   164,    # s8
        0x49:   168,    # s9
        0x4a:   172,    # s10
        0x4b:   176,    # s11
        0x4c:   180,    # s12
        0x4d:   184,    # s13
        0x4e:   188,    # s14
        0x4f:   192,    # s15
        33:     196,    # fpscr
        # (reserved word: 200)
    }

    def __init__(self, parent, thread):
        super(ThreadXThreadContext, self).__init__(parent)
        self._thread = thread
        self._has_fpu = self.core.has_fpu
        if self.core.architecture != CoreArchitecture.ARMv6M:
            # Use the default offsets for this istance
            self._nofpu_register_offsets = self.NOFPU_REGISTER_OFFSETS
        else:
            # Use a copy with the Cortex-M0 specific offsets for this istance
            self._nofpu_register_offsets = self.NOFPU_REGISTER_OFFSETS.copy()
            self._nofpu_register_offsets.update(
                self.NOFPU_REGISTER_OFFSETS_V6M)

    def read_core_registers_raw(self, reg_list):
        reg_list = [index_for_reg(reg) for reg in reg_list]
        reg_vals = []

        isCurrent = self._thread.is_current
        inException = isCurrent and self._parent.read_core_register('ipsr') > 0

        # If this is the current thread and we're not in an exception, just read the live registers.
        if isCurrent and not inException:
            return self._parent.read_core_registers_raw(reg_list)

        # Because of above tests, from now on, inException implies isCurrent;
        # we are generating the thread view for the RTOS thread where the
        # exception occurred; the actual Handler Mode thread view is produced
        # by HandlerModeThread
        if inException:
            # Reasonable to assume PSP is still valid
            sp = self._parent.read_core_register('psp')
        else:
            sp = self._thread.get_stack_pointer()

        # Determine which register offset table to use and the offsets past the saved state.
        hwStacked = 0x20
        swStacked = 0x24
        table = self._nofpu_register_offsets
        if self._has_fpu:
            try:
                if inException and self.core.is_vector_catch():
                    # Vector catch has just occurred, take live LR
                    exceptionLR = self._parent.read_core_register('lr')
                else:
                    # Read stacked exception return LR.
                    offset = self.FPU_REGISTER_OFFSETS[-1]
                    exceptionLR = self._parent.read32(sp + offset)

                # Check bit 4 of the exception LR to determine if FPU registers were stacked.
                if (exceptionLR & EXC_RETURN_EXT_FRAME_MASK) == 0:
                    table = self.FPU_REGISTER_OFFSETS
                    hwStacked = 0x68
                    swStacked = 0x64
            except exceptions.TransferError:
                LOG.debug("Transfer error while reading thread's saved LR")

        for reg in reg_list:

            # Must handle stack pointer specially.
            if reg == 13:
                if inException:
                    reg_vals.append(sp + hwStacked)
                else:
                    reg_vals.append(sp + swStacked + hwStacked)
                continue

            # Look up offset for this register on the stack.
            spOffset = table.get(reg, None)
            if spOffset is None:
                reg_vals.append(self._parent.read_core_register_raw(reg))
                continue
            if inException:
                spOffset -= swStacked

            try:
                if spOffset >= 0:
                    reg_vals.append(self._parent.read32(sp + spOffset))
                else:
                    # Not available - try live one
                    reg_vals.append(self._parent.read_core_register_raw(reg))
            except exceptions.TransferError:
                reg_vals.append(0)

        return reg_vals


class ThreadXThread(TargetThread):
    """@brief A ThreadX task."""

    STATE_NAMES = {
        0: "Ready",
        1: "Completed",
        2: "Terminated",
        3: "Suspended",
        4: "Sleep",
        5: "Queue",
        6: "Semaphore",
        7: "EventFlag",
        8: "BlockMemory",
        9: "ByteMemory",
        10: "IoDriver",
        11: "File",
        12: "TcpIp",
        13: "Mutex",
        14: "PriorityChange",
        99: "Unknown"
    }

    READY = 0
    PRIORITYCHANGE = 14
    UNKNOWN = 99

    def __init__(self, targetContext, provider, base):
        super(ThreadXThread, self).__init__()
        self._target_context = targetContext
        self._provider = provider
        self._base = base
        self._state = self._target_context.read32(
            self._base + THREAD_STATE_OFFSET)
        self._priority = self._target_context.read32(
            self._base + THREAD_PRIORITY_OFFSET)
        self._name = ""
        namePtr = self._target_context.read32(self._base + THREAD_NAME_OFFSET)
        if namePtr != 0:
            self._name = read_c_string(self._target_context, namePtr)
        if len(self._name) == 0:
            self._name = "Unnamed"
        self._thread_context = ThreadXThreadContext(self._target_context, self)

    def get_stack_pointer(self):
        # Get stack pointer saved in thread struct.
        try:
            return self._target_context.read32(self._base + THREAD_STACK_POINTER_OFFSET)
        except exceptions.TransferError:
            LOG.debug("Transfer error while reading thread's stack pointer @ 0x%08x",
                      self._base + THREAD_STACK_POINTER_OFFSET)
            return 0

    def update_info(self):
        try:
            self._priority = self._target_context.read32(
                self._base + THREAD_PRIORITY_OFFSET)
            self._state = self._target_context.read32(
                self._base + THREAD_STATE_OFFSET)
            if not self.READY <= self._state <= self.PRIORITYCHANGE:
                self._state = self.UNKNOWN
        except exceptions.TransferError:
            LOG.debug("Transfer error while reading thread info")

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @property
    def priority(self):
        return self._priority

    @property
    def unique_id(self):
        return self._base

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        # return "%s; Priority %d" % (self.STATE_NAMES[self.state], self.priority)
        return "%s; Priority %d" % (self.STATE_NAMES[self.state], self.priority)

    @property
    def is_current(self):
        return self._provider.get_actual_current_thread_id() == self.unique_id

    @property
    def context(self):
        return self._thread_context

    def __str__(self):
        return "<ThreadXThread@0x%08x id=%x name=%s>" % (id(self), self.unique_id, self.name)

    def __repr__(self):
        return str(self)


class ThreadXThreadProvider(ThreadProvider):
    """@brief Thread provider for ThreadX.

        To successfully initialize, the following ThreadX symbols are needed:
            _tx_thread_created_ptr:     pointer to list of created processes
            _tx_thread_created_count:   count of created processes
            _tx_thread_current_ptr:     current thread
            _tx_thread_system_state:    ThreadX state: initializing, run, interrupt
    """

    # Scheduler not yet up
    TX_INITIALIZE_IN_PROGRESS = 0xF0F0F0F0

    def __init__(self, target):
        super(ThreadXThreadProvider, self).__init__(target)
        self._created_ptr = None
        self._created_cnt = None
        self._current_ptr = None
        self._system_state = None
        self._threads = {}

    def init(self, symbolProvider):
        self._created_ptr = symbolProvider.get_symbol_value(
            '_tx_thread_created_ptr')
        if self._created_ptr is None:
            return False
        LOG.debug("ThreadX: _tx_thread_created_ptr = 0x%08x", self._created_ptr)

        self._created_cnt = symbolProvider.get_symbol_value(
            '_tx_thread_created_count')
        if self._created_cnt is None:
            return False
        LOG.debug("ThreadX: _tx_thread_created_cnt = 0x%08x", self._created_cnt)

        self._current_ptr = symbolProvider.get_symbol_value(
            '_tx_thread_current_ptr')
        if self._current_ptr is None:
            return False
        LOG.debug("ThreadX: _tx_thread_current_ptr = 0x%08x", self._current_ptr)

        self._system_state = symbolProvider.get_symbol_value(
            '_tx_thread_system_state')
        if self._system_state is None:
            return False
        LOG.debug("ThreadX: _tx_thread_system_state = 0x%08x",
                  self._current_ptr)

        self._target.session.subscribe(
            self.event_handler, Target.Event.POST_FLASH_PROGRAM)
        self._target.session.subscribe(
            self.event_handler, Target.Event.POST_RESET)

        return True

    def invalidate(self):
        self._threads = {}

    def event_handler(self, notification):
        # Invalidate threads list if flash is reprogrammed.
        LOG.debug("ThreadX: invalidating threads list: %s" %
                  (repr(notification)))
        self.invalidate()

    def _build_thread_list(self):
        # Read the number of threads.
        threadCount = self._target_context.read32(self._created_cnt)

        # Build up list of all the threads
        allThreads = TargetList(self._target_context, self._created_ptr)
        newThreads = {}
        for threadBase in allThreads:
            try:
                # Reuse existing thread objects if possible.
                if threadBase in self._threads:
                    t = self._threads[threadBase]

                    # Ask the thread object to update its state and priority.
                    t.update_info()
                else:
                    t = ThreadXThread(self._target_context, self, threadBase)
                LOG.debug("Thread 0x%08x (%s)", threadBase, t.name)
                newThreads[t.unique_id] = t
            except exceptions.TransferError:
                LOG.debug(
                    "TransferError while examining thread 0x%08x", threadBase)

        # Is the number of threads correct?
        if len(newThreads) != threadCount:
            LOG.warning("ThreadX: thread count mismatch, %d expected, %d found",
                        threadCount, len(newThreads))

        # Create fake handler mode thread.
        if self._target_context.read_core_register('ipsr') > 0:
            LOG.debug("ThreadX: creating handler mode thread")
            t = HandlerModeThread(self._target_context, self)
            newThreads[t.unique_id] = t

        self._threads = newThreads

    def get_threads(self):
        if not self.is_enabled:
            return []
        self.update_threads()
        return list(self._threads.values())

    def get_thread(self, threadId):
        if not self.is_enabled:
            return None
        self.update_threads()
        return self._threads.get(threadId, None)

    @property
    def is_enabled(self):
        # The _tx_thread_system_state global is used to determine whether
        # the kernel is running. Before the kernel starts, it'll contain
        # TX_INITIALIZE_IN_PROGRESS and possibly TX_INITIALIZE_IN_PROGRESS+1.
        # On cortex-m ports it should otherwise be 0.
        # As it's used in other ports to indicate the interrupt nesting level, it's
        # safer to compare it with TX_INITIALIZE_IN_PROGRESS.
        if self._system_state is None:
            return False
        try:
            return self._target_context.read32(self._system_state) < self.TX_INITIALIZE_IN_PROGRESS
        except exceptions.TransferFaultError:
            LOG.warn("ThreadX: read system state failed, target memory might not be initialized yet.")
            return False

    @property
    def current_thread(self):
        if not self.is_enabled:
            return None
        self.update_threads()
        id = self.get_current_thread_id()
        try:
            return self._threads[id]
        except KeyError:
            return None

    def is_valid_thread_id(self, threadId):
        if not self.is_enabled:
            return False
        self.update_threads()
        return threadId in self._threads

    def get_current_thread_id(self):
        if not self.is_enabled:
            return None
        if self._target_context.read_core_register('ipsr') > 0:
            return HandlerModeThread.UNIQUE_ID
        return self.get_actual_current_thread_id()

    def get_actual_current_thread_id(self):
        if not self.is_enabled:
            return None
        return self._target_context.read32(self._current_ptr)


class ThreadXPlugin(Plugin):
    """@brief Plugin class for ThreadX."""

    def load(self):
        return ThreadXThreadProvider

    @property
    def name(self):
        return "threadx"

    @property
    def description(self):
        return "ThreadX"
